# Copyright (c) Xavier Figueroa
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import datetime
from enum import Enum, auto
from typing import List, Dict, Tuple, Callable, Optional

from spdx import document
from spdx import utils
from spdx.license import LicenseConjunction, LicenseDisjunction
from spdx.package import ExternalPackageRef, PackagePurpose, Package
from spdx.parsers import rdf
from spdx.parsers.builderexceptions import SPDXValueError, CardinalityError, OrderError
from spdx.parsers.loggers import ErrorMessages
from spdx.snippet import Snippet
from spdx.utils import UnKnown, NoAssert

ERROR_MESSAGES = rdf.ERROR_MESSAGES


class BaseParser(object):
    def __init__(self, builder, logger):
        self.builder = builder
        self.logger = logger

    def order_error(self, first_tag, second_tag):
        """
        Helper method for logging an OrderError raised.
        - first_tag: field to be added
        - second_tag: required field
        """
        self.error = True
        msg = "{0} Can not appear before {1}".format(first_tag, second_tag)
        self.logger.log(msg)

    def more_than_one_error(self, field):
        """
        Helper method for logging an CardinalityError raised.
        - field: field/property that has been already defined.
        """
        msg = "More than one {0} defined.".format(field)
        self.logger.log(msg)
        self.error = True

    def value_error(self, key, bad_value):
        """
        Helper method for logging an SPDXValueError raised.
        It reports a value error using ERROR_MESSAGES dict.
        - key: key to use for ERROR_MESSAGES. If not present, a default message is logged
        - bad_value: malformed value
        """
        msg = ERROR_MESSAGES.get(key)
        if msg:
            self.logger.log(msg.format(bad_value))
        else:
            msg = "'{0}' is not a valid value for {1}".format(bad_value, key)
            self.logger.log(msg)
        self.error = True


class CreationInfoParser(BaseParser):
    def __init__(self, builder, logger):
        super(CreationInfoParser, self).__init__(builder, logger)

    def parse_creation_info(self, creation_info):
        """
        Parse Creation Information fields
        - creation_info: Python dict with Creation Information fields in it
        """
        if isinstance(creation_info, dict):
            self.parse_creation_info_comment(creation_info.get("comment"))
            self.parse_creation_info_lic_list_version(
                creation_info.get("licenseListVersion")
            )
            self.parse_creation_info_created(creation_info.get("created"))
            self.parse_creation_info_creators(creation_info.get("creators"))
        else:
            self.value_error("CREATION_INFO_SECTION", creation_info)

    def parse_creation_info_comment(self, comment):
        """
        Parse CreationInfo comment
        - comment: Python str/unicode
        """
        if isinstance(comment, str):
            try:
                return self.builder.set_creation_comment(self.document, comment)
            except CardinalityError:
                self.more_than_one_error("CreationInfo comment")
        elif comment is not None:
            self.value_error("CREATION_COMMENT", comment)

    def parse_creation_info_lic_list_version(self, license_list_version):
        """
        Parse CreationInfo license list version
        - license_list_version: Python str/unicode
        """
        if isinstance(license_list_version, str):
            try:
                return self.builder.set_lics_list_ver(
                    self.document, license_list_version
                )
            except SPDXValueError:
                raise
                self.value_error("LL_VALUE", license_list_version)
            except CardinalityError:
                self.more_than_one_error("CreationInfo licenseListVersion")
        elif license_list_version is not None:
            self.value_error("LL_VALUE", license_list_version)

    def parse_creation_info_created(self, created):
        """
        Parse CreationInfo creation date
        - created: Python str/unicode (ISO-8601 representation of datetime)
        """
        if isinstance(created, str):
            try:
                return self.builder.set_created_date(self.document, created)
            except SPDXValueError:
                self.value_error("CREATED_VALUE", created)
            except CardinalityError:
                self.more_than_one_error("CreationInfo created")
        else:
            self.value_error("CREATED_VALUE", created)

    def parse_creation_info_creators(self, creators):
        """
        Parse CreationInfo creators
        - creators: Python list of creators (str/unicode)
        """
        if isinstance(creators, list):
            for creator in creators:
                if isinstance(creator, str):
                    entity = self.builder.create_entity(self.document, creator)
                    try:
                        self.builder.add_creator(self.document, entity)
                    except SPDXValueError:
                        self.value_error("CREATOR_VALUE", creator)
                else:
                    self.value_error("CREATOR_VALUE", creator)
        else:
            self.value_error("CREATORS_SECTION", creators)


class ExternalDocumentRefsParser(BaseParser):
    def __init__(self, builder, logger):
        super(ExternalDocumentRefsParser, self).__init__(builder, logger)

    def parse_external_document_refs(self, external_document_refs):
        """
        Parse External Document References fields
        - external_document_refs: Python list with External Document References dicts in it
        """
        if isinstance(external_document_refs, list):
            for external_document_ref in external_document_refs:
                if isinstance(external_document_ref, dict):
                    self.parse_ext_doc_ref_id(
                        external_document_ref.get("externalDocumentId")
                    )
                    self.parse_ext_doc_ref_namespace(
                        external_document_ref.get("spdxDocument")
                    )
                    self.parse_ext_doc_ref_chksum(external_document_ref.get("checksum"))
                else:
                    self.value_error("EXT_DOC_REF", external_document_ref)
        elif external_document_refs is not None:
            self.value_error("EXT_DOC_REFS_SECTION", external_document_refs)

    def parse_ext_doc_ref_id(self, ext_doc_ref_id):
        """
        Parse ExternalDocumentReference id
        ext_doc_ref_id: Python str/unicode
        """
        if isinstance(ext_doc_ref_id, str):
            return self.builder.set_ext_doc_id(self.document, ext_doc_ref_id)
        self.value_error("EXT_DOC_REF_ID", ext_doc_ref_id)
        return self.builder.set_ext_doc_id(self.document, "dummy_ext_doc_ref")
        # ext_doc_ref_id is set even if it is None or not string. If weren't, the other attributes
        # would be added to the ex_doc_ref previously added.
        # Another approach is to skip the whole ex_doc_ref itself

    def parse_ext_doc_ref_namespace(self, namespace):
        """
        Parse ExternalDocumentReference namespace
        namespace: Python str/unicode
        """
        if isinstance(namespace, str):
            try:
                return self.builder.set_spdx_doc_uri(self.document, namespace)
            except SPDXValueError:
                self.value_error("EXT_DOC_REF_VALUE", namespace)
        else:
            self.value_error("EXT_DOC_REF_VALUE", namespace)

    def parse_ext_doc_ref_chksum(self, chksum):
        """
        Parse ExternalDocumentReference checksum
        chksum: Python dict('algorithm':str/unicode, 'value':str/unicode)
        """
        if isinstance(chksum, dict):
            value = chksum.get("checksumValue")
            if isinstance(value, str):
                try:
                    return self.builder.set_chksum(self.document, value)
                except SPDXValueError:
                    self.value_error("CHECKSUM_VALUE", value)
            else:
                self.value_error("CHECKSUM_VALUE", value)
        else:
            self.value_error("CHECKSUM_FIELD", chksum)


class LicenseParser(BaseParser):
    def __init__(self, builder, logger):
        super(LicenseParser, self).__init__(builder, logger)

    def parse_extracted_license_info(self, extracted_license_info):
        """
        Parse Extracted Lisence Information fields
        - extracted_license_info: Python list with Extracted Lisence Information dicts in it
        """
        if isinstance(extracted_license_info, list):
            for extracted_license in extracted_license_info:
                if isinstance(extracted_license, dict):
                    if self.parse_ext_lic_id(extracted_license.get("licenseId")):
                        self.parse_ext_lic_name(extracted_license.get("name"))
                        self.parse_ext_lic_comment(extracted_license.get("comment"))
                        self.parse_ext_lic_text(extracted_license.get("extractedText"))
                        self.parse_ext_lic_cross_refs(extracted_license.get("seeAlsos"))
                else:
                    self.value_error("EXTR_LIC", extracted_license)

    def parse_ext_lic_id(self, ext_lic_id):
        """
        Parse ExtractedLicenseInformation id
        ext_lic_id: Python str/unicode
        """
        if isinstance(ext_lic_id, str):
            try:
                return self.builder.set_lic_id(self.document, ext_lic_id)
            except SPDXValueError:
                self.value_error("EXTR_LIC_ID", ext_lic_id)
        else:
            self.value_error("EXTR_LIC_ID", ext_lic_id)

    def parse_ext_lic_name(self, ext_lic_name):
        """
        Parse ExtractedLicenseInformation name
        ext_lic_name: Python str/unicode
        """
        try:
            return self.builder.set_lic_name(self.document, ext_lic_name)
        except SPDXValueError:
            self.value_error("EXTR_LIC_NAME", ext_lic_name)
        except CardinalityError:
            self.more_than_one_error("ExtractedLicense name")
        except OrderError:
            self.order_error("ExtractedLicense name", "ExtractedLicense id")

    def parse_ext_lic_comment(self, ext_lic_comment):
        """
        Parse ExtractedLicenseInformation comment
        ext_lic_comment: Python str/unicode
        """
        if isinstance(ext_lic_comment, str):
            try:
                return self.builder.set_lic_comment(self.document, ext_lic_comment)
            except CardinalityError:
                self.more_than_one_error("ExtractedLicense comment")
            except OrderError:
                self.order_error("ExtractedLicense comment", "ExtractedLicense id")
        elif ext_lic_comment is not None:
            self.value_error("EXTR_LIC_COMMENT", ext_lic_comment)

    def parse_ext_lic_text(self, ext_lic_text):
        """
        Parse ExtractedLicenseInformation text
        ext_lic_text: Python str/unicode
        """
        if isinstance(ext_lic_text, str):
            try:
                return self.builder.set_lic_text(self.document, ext_lic_text)
            except CardinalityError:
                self.more_than_one_error("ExtractedLicense text")
            except OrderError:
                self.order_error("ExtractedLicense text", "ExtractedLicense id")
        else:
            self.value_error("EXTR_LIC_TXT", ext_lic_text)

    def parse_ext_lic_cross_refs(self, cross_refs):
        """
        Parse ExtractedLicenseInformation cross references
        cross_refs: Python list of cross references (str/unicode)
        """
        if isinstance(cross_refs, list):
            for cross_ref in cross_refs:
                if isinstance(cross_ref, str):
                    try:
                        self.builder.add_lic_xref(self.document, cross_ref)
                    except OrderError:
                        self.order_error(
                            "ExtractedLicense cross references", "ExtractedLicense id"
                        )
                else:
                    self.value_error("CROSS_REF", cross_ref)

    def replace_license(self, license_object):
        if isinstance(license_object, LicenseConjunction):
            return LicenseConjunction(
                self.replace_license(license_object.license_1),
                self.replace_license(license_object.license_2),
            )
        elif isinstance(license_object, LicenseDisjunction):
            return LicenseDisjunction(
                self.replace_license(license_object.license_1),
                self.replace_license(license_object.license_2),
            )
        else:
            license_objects = list(
                filter(
                    lambda lic: lic.identifier == license_object.identifier,
                    self.document.extracted_licenses,
                )
            )
            return license_objects[-1] if license_objects else license_object


class AnnotationParser(BaseParser):
    def __init__(self, builder, logger):
        super(AnnotationParser, self).__init__(builder, logger)

    def parse_annotations(self, annotations, spdx_id: str = None):
        """
        Parse Annotation Information fields
        - annotations: Python list with Annotation Information dicts in it
        """
        if isinstance(annotations, list):
            for annotation in annotations:
                if isinstance(annotation, dict):
                    if self.parse_annotation_annotator(annotation.get("annotator")):
                        self.parse_annotation_date(annotation.get("annotationDate"))
                        self.parse_annotation_comment(annotation.get("comment"))
                        self.parse_annotation_type(annotation.get("annotationType"))
                        if annotation.get("SPDXID"):
                            self.parse_annotation_id(annotation.get("SPDXID"))
                        else:
                            self.parse_annotation_id(spdx_id)
                else:
                    self.value_error("ANNOTATION", annotation)

    def parse_annotation_annotator(self, annotator):
        """
        Parse Annotation annotator
        - annotator: Python str/unicode
        """
        if isinstance(annotator, str):
            entity = self.builder.create_entity(self.document, annotator)
            try:
                return self.builder.add_annotator(self.document, entity)
            except SPDXValueError:
                self.value_error("ANNOTATOR_VALUE", annotator)
        else:
            self.value_error("ANNOTATOR_VALUE", annotator)

    def parse_annotation_date(self, date):
        """
        Parse Annotation date
        - date: Python str/unicode (ISO-8601 representation of datetime)
        """
        if isinstance(date, str):
            try:
                return self.builder.add_annotation_date(self.document, date)
            except SPDXValueError:
                self.value_error("ANNOTATION_DATE", date)
            except CardinalityError:
                self.more_than_one_error("Annotation date")
            except OrderError:
                self.order_error("ANNOTATION_DATE", "ANNOTATOR_VALUE")
        else:
            self.value_error("ANNOTATION_DATE", date)

    def parse_annotation_comment(self, comment):
        """
        Parse Annotation comment
        - comment: Python str/unicode
        """
        if isinstance(comment, str):
            try:
                return self.builder.add_annotation_comment(self.document, comment)
            except CardinalityError:
                self.more_than_one_error("Annotation comment")
            except OrderError:
                self.order_error("ANNOTATION_COMMENT", "ANNOTATOR_VALUE")
        else:
            self.value_error("ANNOTATION_COMMENT", comment)

    def parse_annotation_type(self, annotation_type):
        """
        Parse Annotation type
        - annotation_type: Python str/unicode (REVIEW or OTHER)
        """
        if isinstance(annotation_type, str):
            try:
                return self.builder.add_annotation_type(self.document, annotation_type)
            except SPDXValueError:
                self.value_error("ANNOTATION_TYPE", annotation_type)
            except CardinalityError:
                self.more_than_one_error("ANNOTATION_TYPE")
            except OrderError:
                self.order_error("ANNOTATION_TYPE", "ANNOTATOR_VALUE")
        else:
            self.value_error("ANNOTATION_TYPE", annotation_type)

    def parse_annotation_id(self, annotation_id):
        """
        Parse Annotation id
        - annotation_id: Python str/unicode
        """
        if isinstance(annotation_id, str):
            try:
                return self.builder.set_annotation_spdx_id(self.document, annotation_id)
            except CardinalityError:
                self.more_than_one_error("ANNOTATION_ID")
            except OrderError:
                self.order_error("ANNOTATION_ID", "ANNOTATOR_VALUE")
        else:
            self.value_error("ANNOTATION_ID", annotation_id)


class RelationshipParser(BaseParser):
    def __init__(self, builder, logger):
        super(RelationshipParser, self).__init__(builder, logger)

    def parse_relationships(self, relationships):
        """
        Parse Relationship Information fields
        - relationships: Python list with Relationship Information dicts in it
        """
        if isinstance(relationships, list):
            for relationship in relationships:
                if isinstance(relationship, dict):
                    if self.parse_relationship(
                        relationship.get("spdxElementId"),
                        relationship.get("relationshipType"),
                        relationship.get("relatedSpdxElement"),
                    ):
                        self.parse_relationship_comment(relationship.get("comment"))
                else:
                    self.value_error("RELATIONSHIP", relationship)

    def parse_relationship(self, spdxelementid, relationshiptype, relatedspdxelement):
        """
        Parse Relationshiptype, spdxElementId and relatedSpdxElement
        - relationship: Python str/unicode
        """
        if not isinstance(relationshiptype, str):
            self.value_error("RELATIONSHIP_VALUE", relationshiptype)
            return
        if not isinstance(spdxelementid, str):
            self.value_error("SPDXELEMENTID", spdxelementid)
            return
        if not isinstance(relatedspdxelement, str):
            self.value_error("RELATEDSPDXELEMENT", relatedspdxelement)
            return
        relate = spdxelementid + " " + relationshiptype + " " + relatedspdxelement
        try:
            return self.builder.add_relationship(self.document, relate)
        except SPDXValueError:
            self.value_error("RELATIONSHIP_VALUE", relate)


    def parse_relationship_comment(self, relationship_comment):
        """
        Parse relationship comment
        - relationship_comment: Python str/unicode
        """
        if isinstance(relationship_comment, str):
            try:
                return self.builder.add_relationship_comment(
                    self.document, relationship_comment
                )
            except CardinalityError:
                self.more_than_one_error("RELATIONSHIP_COMMENT")
            except OrderError:
                self.order_error("RELATIONSHIP_COMMENT", "RELATIONSHIP")
        elif relationship_comment is not None:
            self.value_error("RELATIONSHIP_COMMENT", relationship_comment)


class RangeType(Enum):
    BYTE = auto()
    LINE = auto()


class SnippetParser(BaseParser):
    def __init__(self, builder, logger):
        super(SnippetParser, self).__init__(builder, logger)

    @property
    def snippet(self) -> Snippet:
        return self.document.snippet[-1]

    def parse_snippets(self, snippets):
        """
        Parse Snippet Information fields
        - snippets: Python list with Snippet Information dicts in it
        """
        if isinstance(snippets, list):
            for snippet in snippets:
                if isinstance(snippet, dict):
                    if self.parse_snippet_id(snippet.get("SPDXID")):
                        self.parse_snippet_name(snippet.get("name"))
                        self.parse_snippet_comment(snippet.get("comment"))
                        self.parse_snippet_copyright(snippet.get("copyrightText"))
                        self.parse_snippet_license_comment(
                            snippet.get("licenseComments")
                        )
                        self.parse_snippet_file_spdxid(snippet.get("snippetFromFile"))
                        self.parse_snippet_concluded_license(
                            snippet.get("licenseConcluded")
                        )
                        self.parse_snippet_attribution_text(
                            snippet.get("attributionTexts")
                        )
                        self.parse_snippet_license_info_from_snippet(
                            snippet.get("licenseInfoInSnippets")
                        )
                        self.parse_annotations(snippet.get("annotations"), spdx_id=snippet.get("SPDXID"))
                        self.parse_snippet_ranges(snippet.get("ranges"))
                else:
                    self.value_error("SNIPPET", snippet)

    def parse_snippet_id(self, snippet_id):
        """
        Parse Snippet id
        - snippet_id: Python str/unicode
        """
        if isinstance(snippet_id, str):
            try:
                return self.builder.create_snippet(self.document, snippet_id)
            except SPDXValueError:
                self.value_error("SNIPPET_SPDX_ID_VALUE", snippet_id)
        else:
            self.value_error("SNIPPET_SPDX_ID_VALUE", snippet_id)

    def parse_snippet_name(self, snippet_name):
        """
        Parse Snippet name
        - snippet_name: Python str/unicode
        """
        if isinstance(snippet_name, str):
            try:
                return self.builder.set_snippet_name(self.document, snippet_name)
            except CardinalityError:
                self.more_than_one_error("SNIPPET_NAME")
        elif snippet_name is not None:
            self.value_error("SNIPPET_NAME", snippet_name)

    def parse_snippet_comment(self, snippet_comment):
        """
        Parse Snippet comment
        - snippet_comment: Python str/unicode
        """
        if isinstance(snippet_comment, str):
            try:
                return self.builder.set_snippet_comment(self.document, snippet_comment)
            except CardinalityError:
                self.more_than_one_error("SNIPPET_COMMENT")
        elif snippet_comment is not None:
            self.value_error("SNIPPET_COMMENT", snippet_comment)

    def parse_snippet_attribution_text(self, snippet_attribution_texts):
        """
        Parse Snippet attribution texts
        - snippet_attribution_texts: list in yaml, json and string in xml format
        """
        if isinstance(snippet_attribution_texts, list) or isinstance(
            snippet_attribution_texts, str
        ):
            for snippet_attribution_text in snippet_attribution_texts:
                try:
                    return self.builder.set_snippet_attribution_text(
                        self.document, snippet_attribution_texts
                    )
                except CardinalityError:
                    self.more_than_one_error("SNIPPET_ATTRIBUTION_TEXT")
                except OrderError:
                    self.order_error("SNIPPET_ATTRIBUTION_TEXT", "SNIPPET_NAME")
            else:
                self.value_error("SNIPPET_ATTRIBUTION_TEXT", snippet_attribution_texts)

    def parse_snippet_copyright(self, copyright_text):
        """
        Parse Snippet copyright text
        - copyright_text: Python str/unicode
        """
        if isinstance(copyright_text, str):
            try:
                return self.builder.set_snippet_copyright(self.document, copyright_text)
            except CardinalityError:
                self.more_than_one_error("SNIPPET_COPYRIGHT")
        elif copyright_text is not None:
            self.value_error("SNIPPET_COPYRIGHT", copyright_text)

    def parse_snippet_license_comment(self, license_comment):
        """
        Parse Snippet license comment
        - license_comment: Python str/unicode
        """
        if isinstance(license_comment, str):
            try:
                return self.builder.set_snippet_lic_comment(
                    self.document, license_comment
                )
            except CardinalityError:
                self.more_than_one_error("SNIPPET_LIC_COMMENTS")
        elif license_comment is not None:
            self.value_error("SNIPPET_LIC_COMMENTS", license_comment)

    def parse_snippet_file_spdxid(self, file_spdxid):
        """
        Parse Snippet file id
        - file_spdxid: Python str/unicode
        """
        if isinstance(file_spdxid, str):
            try:
                return self.builder.set_snip_from_file_spdxid(
                    self.document, file_spdxid
                )
            except SPDXValueError:
                self.value_error("SNIPPET_FILE_ID", file_spdxid)
            except CardinalityError:
                self.more_than_one_error("SNIPPET_FILE_ID")
        else:
            self.value_error("SNIPPET_FILE_ID", file_spdxid)

    def parse_snippet_concluded_license(self, concluded_license):
        """
        Parse Snippet concluded license
        - concluded_license: Python str/unicode
        """
        if isinstance(concluded_license, str):
            lic_parser = utils.LicenseListParser()
            lic_parser.build(write_tables=0, debug=0)
            license_object = self.replace_license(lic_parser.parse(concluded_license))
            try:
                return self.builder.set_snip_concluded_license(
                    self.document, license_object
                )
            except SPDXValueError:
                self.value_error("SNIPPET_CONCLUDED_LICENSE", concluded_license)
            except CardinalityError:
                self.more_than_one_error("SNIPPET_CONCLUDED_LICENSE")
        elif concluded_license is not None:
            self.value_error("SNIPPET_CONCLUDED_LICENSE", concluded_license)

    def parse_snippet_license_info_from_snippet(self, license_info_from_snippet):
        """
        Parse Snippet license information from snippet
        - license_info_from_snippet: Python list of licenses information from snippet (str/unicode)
        """
        if isinstance(license_info_from_snippet, list):
            for lic_in_snippet in license_info_from_snippet:
                if isinstance(lic_in_snippet, str):
                    lic_parser = utils.LicenseListParser()
                    lic_parser.build(write_tables=0, debug=0)
                    license_object = self.replace_license(
                        lic_parser.parse(lic_in_snippet)
                    )
                    try:
                        self.builder.set_snippet_lics_info(
                            self.document, license_object
                        )
                    except SPDXValueError:
                        self.value_error("SNIPPET_LIC_INFO", lic_in_snippet)
                else:
                    self.value_error("SNIPPET_LIC_INFO", lic_in_snippet)
        elif license_info_from_snippet is not None:
            self.value_error("SNIPPET_LIC_INFO_FIELD", license_info_from_snippet)

    def parse_snippet_ranges(self, ranges_from_snippet: List[Dict]) -> None:
        """
        Parse ranges (byte range and optional line range) from snippet
        - ranges_from_snippet; Python list of dict
        """
        if not isinstance(ranges_from_snippet, list):
            self.value_error("SNIPPET_RANGES", ranges_from_snippet)
            return

        for range_dict in ranges_from_snippet:
            try:
                range_type = self.validate_range_and_get_type(range_dict)
                start_end_tuple: Tuple[int, int] = SnippetParser.get_start_end_tuple(range_dict, range_type)
            except SPDXValueError:
                self.value_error("SNIPPET_RANGE", range_dict)
                return

            if range_type == RangeType.BYTE:
                self.snippet.byte_range = start_end_tuple
            elif range_type == RangeType.LINE:
                self.snippet.line_range = start_end_tuple

    @staticmethod
    def get_start_end_tuple(range_dict: Dict, range_type: RangeType) -> Tuple[int, int]:
        end_pointer = range_dict["endPointer"]
        start_pointer = range_dict["startPointer"]
        if range_type == RangeType.BYTE:
            start = int(start_pointer["offset"])
            end = int(end_pointer["offset"])
        else:
            start = int(start_pointer["lineNumber"])
            end = int(end_pointer["lineNumber"])
        if start > end:
            raise SPDXValueError("Snippet::ranges")

        return start, end

    def validate_range_and_get_type(self, range_dict: Dict) -> RangeType:
        if ("startPointer" not in range_dict) or ("endPointer" not in range_dict):
            raise SPDXValueError("Snippet::ranges")
        start_pointer_type = self.validate_pointer_and_get_type(range_dict["startPointer"])
        end_pointer_type = self.validate_pointer_and_get_type(range_dict["endPointer"])
        if start_pointer_type != end_pointer_type:
            raise SPDXValueError("Snippet::ranges")
        return start_pointer_type

    def validate_pointer_and_get_type(self, pointer: Dict) -> RangeType:
        if self.snippet.snip_from_file_spdxid != pointer["reference"]:
            raise SPDXValueError("Snippet::ranges")
        if ("offset" in pointer and "lineNumber" in pointer) or (
                "offset" not in pointer and "lineNumber" not in pointer):
            raise SPDXValueError("Snippet::ranges")

        return RangeType.BYTE if "offset" in pointer else RangeType.LINE


class ReviewParser(BaseParser):
    def __init__(self, builder, logger):
        super(ReviewParser, self).__init__(builder, logger)

    def parse_reviews(self, reviews):
        """
        Parse Review Information fields
        - reviews: Python list with Review Information dicts in it
        """
        if isinstance(reviews, list):
            for review in reviews:
                if isinstance(review, dict):
                    if self.parse_review_reviewer(review.get("reviewer")):
                        self.parse_review_date(review.get("reviewDate"))
                        self.parse_review_comment(review.get("comment"))
                else:
                    self.value_error("REVIEW", review)

    def parse_review_reviewer(self, reviewer):
        """
        Parse Review reviewer
        - reviewer: Python str/unicode
        """
        if isinstance(reviewer, str):
            entity = self.builder.create_entity(self.document, reviewer)
            try:
                return self.builder.add_reviewer(self.document, entity)
            except SPDXValueError:
                self.value_error("REVIEWER_VALUE", reviewer)
        else:
            self.value_error("REVIEWER_VALUE", reviewer)

    def parse_review_date(self, review_date):
        """
        Parse Review date
        - review_date: Python str/unicode (ISO-8601 representation of datetime)
        """
        if isinstance(review_date, str):
            try:
                return self.builder.add_review_date(self.document, review_date)
            except SPDXValueError:
                self.value_error("REVIEW_DATE", review_date)
            except CardinalityError:
                self.more_than_one_error("REVIEW_DATE")
            except OrderError:
                self.order_error("REVIEW_DATE", "REVIEWER_VALUE")
        else:
            self.value_error("REVIEW_DATE", review_date)

    def parse_review_comment(self, review_comment):
        """
        Parse Review comment
        - review_comment: Python str/unicode
        """
        if isinstance(review_comment, str):
            try:
                return self.builder.add_review_comment(self.document, review_comment)
            except CardinalityError:
                self.more_than_one_error("REVIEW_COMMENT")
            except OrderError:
                self.order_error("REVIEW_COMMENT", "REVIEWER_VALUE")
        elif review_comment is not None:
            self.value_error("REVIEW_COMMENT", review_comment)


class FileParser(BaseParser):
    def __init__(self, builder, logger):
        super(FileParser, self).__init__(builder, logger)

    def parse_file(self, file):
        """
        Parse File Information fields
        - file: Python dict with File Information fields in it
        """
        if isinstance(file, dict):
            self.parse_file_name(file.get("fileName"))
            self.parse_file_id(file.get("SPDXID"))
            self.parse_file_types(file.get("fileTypes"))
            self.parse_file_concluded_license(file.get("licenseConcluded"))
            self.parse_file_license_info_in_files(file.get("licenseInfoInFiles"))
            self.parse_file_license_comments(file.get("licenseComments"))
            self.parse_file_copyright_text(file.get("copyrightText"))
            self.parse_file_artifacts(file.get("artifactOf"))
            self.parse_file_comment(file.get("comment"))
            self.parse_file_notice_text(file.get("noticeText"))
            self.parse_file_contributors(file.get("fileContributors"))
            self.parse_file_attribution_text(file.get("attributionTexts"))
            self.parse_file_dependencies(file.get("fileDependencies"))
            self.parse_annotations(file.get("annotations"), spdx_id=file.get("SPDXID"))
            self.parse_file_checksums(file.get("checksums"))
        else:
            self.value_error("FILE", file)

    def parse_file_name(self, file_name):
        """
        Parse File name
        - file_name: Python str/unicode
        """
        if isinstance(file_name, str):
            return self.builder.set_file_name(self.document, file_name)
        self.value_error("FILE_NAME", file_name)
        return self.builder.set_file_name(self.document, "dummy_file")
        # file_name is set even if it is None or not string. If weren't, the other attributes
        # would be added to the file previously added.
        # Another approach is to skip the whole file itself

    def parse_file_id(self, file_id):
        """
        Parse File id
        - file_id: Python str/unicode
        """
        if isinstance(file_id, str):
            try:
                return self.builder.set_file_spdx_id(self.document, file_id)
            except SPDXValueError:
                self.value_error("FILE_ID", file_id)
            except CardinalityError:
                self.more_than_one_error("FILE_ID")
            except OrderError:
                self.order_error("FILE_ID", "FILE_NAME")
        else:
            self.value_error("FILE_ID", file_id)

    def parse_file_types(self, file_types):
        """
        Parse File types
        - file_types: Python list of file types (str/unicode: fileType_archive, fileType_binary, fileType_source or
        fileType_other)
        """
        if isinstance(file_types, list):  # file_types is an array in JSON examples...
            for file_type in file_types:
                self.parse_file_type(file_type)
        # ...but file.File allows only one type at the moment.
        elif isinstance(file_types, str):
            return self.parse_file_type(file_types)
        elif file_types is not None:
            self.value_error("FILE_TYPES", file_types)

    def parse_file_type(self, file_type):
        """
        Parse File type
        - file_type: Python str/unicode (fileType_archive, fileType_binary, fileType_source or fileType_other)
        """
        if isinstance(file_type, str):
            try:
                return self.builder.set_file_type(self.document, file_type)
            except SPDXValueError:
                self.value_error("FILE_TYPE", file_type)
            except OrderError:
                self.order_error("FILE_TYPE", "FILE_NAME")
        else:
            self.value_error("FILE_TYPE", file_type)

    def parse_file_concluded_license(self, concluded_license):
        """
        Parse File concluded license
        - concluded_license: Python str/unicode
        """
        if isinstance(concluded_license, str):
            lic_parser = utils.LicenseListParser()
            lic_parser.build(write_tables=0, debug=0)
            license_object = self.replace_license(lic_parser.parse(concluded_license))
            try:
                return self.builder.set_concluded_license(self.document, license_object)
            except SPDXValueError:
                self.value_error("FILE_SINGLE_LICS", concluded_license)
            except CardinalityError:
                self.more_than_one_error("FILE_SINGLE_LICS")
            except OrderError:
                self.order_error("FILE_SINGLE_LICS", "FILE_NAME")
        elif concluded_license is not None:
            self.value_error("FILE_SINGLE_LICS", concluded_license)

    def parse_file_license_info_in_files(self, license_info_in_files):
        """
        Parse File license information from files
        - license_info_from_files: Python list of licenses information from files (str/unicode)
        """
        if isinstance(license_info_in_files, list):
            for license_info_in_file in license_info_in_files:
                if isinstance(license_info_in_file, str):
                    lic_parser = utils.LicenseListParser()
                    lic_parser.build(write_tables=0, debug=0)
                    license_object = self.replace_license(
                        lic_parser.parse(license_info_in_file)
                    )
                    try:
                        self.builder.set_file_license_in_file(
                            self.document, license_object
                        )
                    except SPDXValueError:
                        self.value_error("FILE_LIC_IN_FILES", license_info_in_file)
                    except OrderError:
                        self.order_error("FILE_LIC_IN_FILES", "FILE_NAME")
                else:
                    self.value_error("FILE_LIC_IN_FILES", license_info_in_file)
        elif license_info_in_files is not None:
            self.value_error("FILE_LIC_IN_FILES_FIELD", license_info_in_files)

    def parse_file_license_comments(self, license_comments):
        """
        Parse File license comments
        - license_comments: Python str/unicode
        """
        if isinstance(license_comments, str):
            try:
                return self.builder.set_file_license_comment(
                    self.document, license_comments
                )
            except CardinalityError:
                self.more_than_one_error("FILE_LIC_COMMENTS")
            except OrderError:
                self.order_error("FILE_LIC_COMMENTS", "FILE_NAME")
        elif license_comments is not None:
            self.value_error("FILE_LIC_COMMENTS", license_comments)

    def parse_file_attribution_text(self, file_attribution_texts):
        """
        Parse File attribution texts
        - file_attribution_texts: list in yaml, json and string in xml format
        """
        if isinstance(file_attribution_texts, list):
            for file_attribution_text in file_attribution_texts:
                try:
                    return self.builder.set_file_attribution_text(
                        self.document, file_attribution_text
                    )
                except CardinalityError:
                    self.more_than_one_error("FILE_ATTRIBUTION_TEXT")
                except OrderError:
                    self.order_error("FILE_ATTRIBUTION_TEXT", "FILE_NAME")
            else:
                self.value_error("FILE_ATTRIBUTION_TEXT", file_attribution_texts)
        elif isinstance(file_attribution_texts, str):
            try:
                return self.builder.set_file_attribution_text(
                    self.document, file_attribution_texts
                )
            except CardinalityError:
                self.more_than_one_error("FILE_ATTRIBUTION_TEXT")
            except OrderError:
                self.order_error("FILE_ATTRIBUTION_TEXT", "FILE_NAME")

    def parse_file_copyright_text(self, copyright_text):
        """
        Parse File copyright text
        - copyright_text: Python str/unicode
        """
        if isinstance(copyright_text, str):
            try:
                return self.builder.set_file_copyright(self.document, copyright_text)
            except CardinalityError:
                self.more_than_one_error("FILE_COPYRIGHT_TEXT")
            except OrderError:
                self.order_error("FILE_COPYRIGHT_TEXT", "FILE_NAME")
        elif copyright_text is not None:
            self.value_error("FILE_COPYRIGHT_TEXT", copyright_text)

    def parse_file_artifacts(self, file_artifacts):
        """
        Parse File artifacts
        - file_artifacts: Python list of dict('name':str/unicode, 'homePage':str/unicode, 'projectUri':str/unicode)
        """
        if isinstance(file_artifacts, list):
            for artifact in file_artifacts:
                if isinstance(artifact, dict):
                    self.builder.set_file_atrificat_of_project(
                        self.document, "name", artifact.get("name", UnKnown())
                    )
                    self.builder.set_file_atrificat_of_project(
                        self.document, "home", artifact.get("homePage", UnKnown())
                    )
                    self.builder.set_file_atrificat_of_project(
                        self.document, "uri", artifact.get("projectUri", UnKnown())
                    )
                    return True
                else:
                    self.value_error("ARTIFACT_OF_VALUE", artifact)
        elif file_artifacts is not None:
            self.value_error("ARTIFACT_OF_FIELD", file_artifacts)

    def parse_file_comment(self, file_comment):
        """
        Parse File comment
        - file_comment: Python str/unicode
        """
        if isinstance(file_comment, str):
            try:
                return self.builder.set_file_comment(self.document, file_comment)
            except CardinalityError:
                self.more_than_one_error("FILE_COMMENT")
            except OrderError:
                self.order_error("FILE_COMMENT", "FILE_NAME")
        elif file_comment is not None:
            self.value_error("FILE_COMMENT", file_comment)

    def parse_file_notice_text(self, notice_text):
        """
        Parse File notice text
        - notice_text: Python str/unicode
        """
        if isinstance(notice_text, str):
            try:
                return self.builder.set_file_notice(self.document, notice_text)
            except CardinalityError:
                self.more_than_one_error("FILE_NOTICE_TEXT")
            except OrderError:
                self.order_error("FILE_NOTICE_TEXT", "FILE_NAME")
        elif notice_text is not None:
            self.value_error("FILE_NOTICE_TEXT", notice_text)

    def parse_file_contributors(self, file_contributors):
        """
        Parse File contributors
        - file_contributors: Python list of contributors (str/unicode)
        """
        if isinstance(file_contributors, list):
            for contributor in file_contributors:
                if isinstance(contributor, str):
                    try:
                        self.builder.add_file_contribution(self.document, contributor)
                    except OrderError:
                        self.order_error("FILE_CONTRIBUTOR", "FILE_NAME")
                else:
                    self.value_error("FILE_CONTRIBUTOR", contributor)
        elif file_contributors is not None:
            self.value_error("FILE_CONTRIBUTORS", file_contributors)

    def parse_file_dependencies(self, file_dependencies):
        """
        Parse File dependencies
        - file_dependencies: Python list of dependencies (str/unicode or file dict as in FileParser.parse_file)
        """
        if isinstance(file_dependencies, list):
            for dependency in file_dependencies:
                dependency = self._handle_file_dependency(dependency)
                if isinstance(dependency, str):
                    try:
                        self.builder.add_file_dep(self.document, dependency)
                    except OrderError:
                        self.order_error("FILE_DEPENDENCY", "FILE_NAME")
                else:
                    self.value_error("FILE_DEPENDENCY", dependency)
        elif file_dependencies is not None:
            self.value_error("FILE_DEPENDENCIES", file_dependencies)

    def _handle_file_dependency(self, file_dependency):
        """
        Helper method that handles file-like dependency
        - file_dependency: Python dict as in FileParser.parse_file
        return: file name (str/unicode) or None
        """
        if isinstance(file_dependency, dict):
            filelike_dependency = file_dependency.get("File")
            if isinstance(filelike_dependency, dict):
                return filelike_dependency.get("name")
            return None
        return None

    def parse_file_checksums(self, file_checksums: List[Dict]) -> Optional[bool]:
        """
        Parse File checksums
        - file_checksums: Python List
        """
        if isinstance(file_checksums, list):
            for checksum in file_checksums:
                self.builder.set_file_checksum(self.document, checksum)
            return True
        if isinstance(file_checksums, str):
            # kept for backwards compatibility
            try:
                return self.builder.set_file_checksum(self.document, file_checksums)
            except CardinalityError:
                self.more_than_one_error("FILE_CHECKSUM")
            except OrderError:
                self.order_error("FILE_CHECKSUM", "FILE_NAME")
        else:
            self.value_error("FILE_CHECKSUM", file_checksums)

    def parse_files(self, files: List[Dict]) -> None:
        if files is None:
            return
        if isinstance(files, list):
            for file in files:
                self.parse_file(file)
        else:
            self.value_error("FILES", files)



class PackageParser(BaseParser):
    def __init__(self, builder, logger):
        super(PackageParser, self).__init__(builder, logger)

    @property
    def package(self):
        # current package being parsed is the last one
        return self.document.packages[-1]

    def parse_package(self, package: Package, method_to_parse_relationship: Callable):
        """
        Parse Package Information fields
        - package: Python dict with Package Information fields in it
        """
        if isinstance(package, dict):
            # The builder has the notion of current package, here, we force to start a new one
            self.builder.reset_package()
            self.parse_pkg_name(package.get("name"))
            self.parse_pkg_id(package.get("SPDXID"))
            self.parse_pkg_files_analyzed(package.get("filesAnalyzed"))
            self.parse_pkg_version(package.get("versionInfo"))
            self.parse_pkg_file_name(package.get("packageFileName"))
            self.parse_pkg_supplier(package.get("supplier"))
            self.parse_pkg_originator(package.get("originator"))
            self.parse_pkg_down_location(package.get("downloadLocation"))
            self.parse_pkg_verif_code_field(package.get("packageVerificationCode"))
            self.parse_pkg_homepage(package.get("homepage"))
            self.parse_pkg_source_info(package.get("sourceInfo"))
            self.parse_pkg_concluded_license(package.get("licenseConcluded"))
            self.parse_pkg_license_info_from_files(package.get("licenseInfoFromFiles"))
            self.parse_pkg_declared_license(package.get("licenseDeclared"))
            self.parse_pkg_license_comment(package.get("licenseComments"))
            self.parse_pkg_copyright_text(package.get("copyrightText"))
            self.parse_pkg_summary(package.get("summary"))
            self.parse_pkg_comment(package.get("comment"))
            self.parse_pkg_description(package.get("description"))
            self.parse_annotations(package.get("annotations"), spdx_id=package.get("SPDXID"))
            self.parse_pkg_attribution_text(package.get("attributionTexts"))
            self.parse_pkg_files(package.get("hasFiles"), method_to_parse_relationship)
            self.parse_pkg_checksums(package.get("checksums"))
            self.parse_package_external_refs(package.get("externalRefs"))
            self.parse_primary_package_purpose(package.get("primaryPackagePurpose"))
            self.parse_release_date(package.get("releaseDate"))
            self.parse_built_date(package.get("builtDate"))
            self.parse_valid_until_date(package.get("validUntilDate"))
        else:
            self.value_error("PACKAGE", package)

    def parse_pkg_name(self, pkg_name):
        """
        Parse Package name
        - pkg_name: Python str/unicode
        """
        if isinstance(pkg_name, str):
            return self.builder.create_package(self.document, pkg_name)
        self.value_error("PKG_NAME", pkg_name)
        return self.builder.create_package(self.document, "dummy_package")
        # pkg_name is set even if it is None or not string. If weren't, the other attributes
        # would be added to the package previously added.
        # Another approach is to skip the whole package itself

    def parse_pkg_id(self, pkg_id):
        """
        Parse Package id
        - pkg_id: Python str/unicode
        """
        if isinstance(pkg_id, str):
            try:
                return self.builder.set_pkg_spdx_id(self.document, pkg_id)
            except SPDXValueError:
                self.value_error("PKG_ID", pkg_id)
            except CardinalityError:
                self.more_than_one_error("PKG_ID")
        else:
            self.value_error("PKG_ID", pkg_id)

    def parse_pkg_version(self, pkg_version):
        """
        Parse Package version
        - pkg_name: Python str/unicode
        """
        if isinstance(pkg_version, str):
            try:
                return self.builder.set_pkg_vers(self.document, pkg_version)
            except CardinalityError:
                self.more_than_one_error("PKG_VERSION")
            except OrderError:
                self.order_error("PKG_VERSION", "PKG_NAME")
        elif pkg_version is not None:
            self.value_error("PKG_VERSION", pkg_version)

    def parse_pkg_file_name(self, pkg_file_name):
        """
        Parse Package file name
        - pkg_file_name: Python str/unicode
        """
        if isinstance(pkg_file_name, str):
            try:
                return self.builder.set_pkg_file_name(self.document, pkg_file_name)
            except CardinalityError:
                self.more_than_one_error("PKG_FILE_NAME")
            except OrderError:
                self.order_error("PKG_FILE_NAME", "PKG_NAME")
        elif pkg_file_name is not None:
            self.value_error("PKG_FILE_NAME", pkg_file_name)

    def parse_pkg_supplier(self, pkg_supplier):
        """
        Parse Package supplier
        - pkg_supplier: Python str/unicode
        """
        if isinstance(pkg_supplier, str):
            if pkg_supplier == "NOASSERTION":
                return self.builder.set_pkg_supplier(self.document, NoAssert())
            entity = self.builder.create_entity(self.document, pkg_supplier)
            try:
                return self.builder.set_pkg_supplier(self.document, entity)
            except SPDXValueError:
                self.value_error("PKG_SUPPL_VALUE", pkg_supplier)
            except CardinalityError:
                self.more_than_one_error("PKG_SUPPL_VALUE")
            except OrderError:
                self.order_error("PKG_SUPPL_VALUE", "PKG_NAME")
        elif pkg_supplier is not None:
            self.value_error("PKG_SUPPL_VALUE", pkg_supplier)

    def parse_pkg_originator(self, pkg_originator):
        """
        Parse Package originator
        - pkg_originator: Python str/unicode
        """
        if isinstance(pkg_originator, str):
            if pkg_originator == "NOASSERTION":
                return self.builder.set_pkg_originator(self.document, NoAssert())
            entity = self.builder.create_entity(self.document, pkg_originator)
            try:
                return self.builder.set_pkg_originator(self.document, entity)
            except SPDXValueError:
                self.value_error("PKG_ORIGINATOR_VALUE", pkg_originator)
            except CardinalityError:
                self.more_than_one_error("PKG_ORIGINATOR_VALUE")
            except OrderError:
                self.order_error("PKG_ORIGINATOR_VALUE", "PKG_NAME")
        elif pkg_originator is not None:
            self.value_error("PKG_ORIGINATOR_VALUE", pkg_originator)

    def parse_pkg_down_location(self, pkg_down_location):
        """
        Parse Package download location
        - pkg_down_location: Python str/unicode
        """
        if isinstance(pkg_down_location, str):
            try:
                return self.builder.set_pkg_down_location(
                    self.document, pkg_down_location
                )
            except CardinalityError:
                self.more_than_one_error("PKG_DOWN_LOC")
            except OrderError:
                self.order_error("PKG_DOWN_LOC", "PKG_NAME")
        else:
            self.value_error("PKG_DOWN_LOC", pkg_down_location)

    def parse_pkg_files_analyzed(self, pkg_files_analyzed):
        """
        Parse Package files analyzed
        - pkg_files_analyzed: Python boolean
        """
        # Files Analyzed optional
        if pkg_files_analyzed is None:
            return

        # For XML, this is a string, not a boolean.
        # xmltodict doesn't do this translation for us, so we do it here.
        if isinstance(pkg_files_analyzed, str):
            if pkg_files_analyzed.lower() in ['true', '1']:
                pkg_files_analyzed = True
            elif pkg_files_analyzed.lower() in ['false', '0']:
                pkg_files_analyzed = False

        if isinstance(pkg_files_analyzed, bool):
            try:
                return self.builder.set_pkg_files_analyzed(
                    self.document, pkg_files_analyzed
                )
            except CardinalityError:
                self.more_than_one_error("PKG_FILES_ANALYZED")
        else:
            self.value_error("PKG_FILES_ANALYZED", pkg_files_analyzed)

    def parse_pkg_verif_code_field(self, pkg_verif_code_field):
        """
        Parse Package verification code dict
        - pkg_verif_code_field: Python dict('value':str/unicode, 'excludedFilesNames':list)
        """
        if not self.package.are_files_analyzed:
            if pkg_verif_code_field is not None:
                self.value_error("PKG_VERIF_CODE_FIELD", pkg_verif_code_field)
            return

        if isinstance(pkg_verif_code_field, dict):
            self.parse_pkg_verif_exc_files(
                pkg_verif_code_field.get("packageVerificationCodeExcludedFiles")
            )
            return self.parse_pkg_verif_code(pkg_verif_code_field.get("packageVerificationCodeValue"))
        elif pkg_verif_code_field is not None:
            self.value_error("PKG_VERIF_CODE_FIELD", pkg_verif_code_field)

    def parse_pkg_verif_code(self, pkg_verif_code):
        """
        Parse Package verification code value
        - pkg_verif_code: Python str/unicode
        """
        if not self.package.are_files_analyzed:
            if pkg_verif_code is not None:
                self.value_error("PKG_VERIF_CODE", pkg_verif_code)
            return

        if isinstance(pkg_verif_code, str):
            try:
                return self.builder.set_pkg_verif_code(self.document, pkg_verif_code)
            except CardinalityError:
                self.more_than_one_error("PKG_VERIF_CODE")
            except OrderError:
                self.order_error("PKG_VERIF_CODE", "PKG_NAME")
        else:
            self.value_error("PKG_VERIF_CODE", pkg_verif_code)

    def parse_pkg_verif_exc_files(self, pkg_verif_exc_files):
        """
        Parse Package files excluded from verification code
        - pkg_verif_exc_files: Python list of files excluded (str/unicode)
        """
        if isinstance(pkg_verif_exc_files, list):
            for pkg_verif_exc_file in pkg_verif_exc_files:
                if isinstance(pkg_verif_exc_file, str):
                    try:
                        self.builder.set_pkg_excl_file(
                            self.document, pkg_verif_exc_file
                        )
                    except OrderError:
                        self.order_error("PKG_VERIF_EXC_FILE", "PKG_NAME")
                else:
                    self.value_error("PKG_VERIF_EXC_FILE", pkg_verif_exc_file)
        elif pkg_verif_exc_files is not None:
            self.value_error("PKG_VERIF_EXC_FILE_FIELD", pkg_verif_exc_files)

    def parse_pkg_homepage(self, pkg_homepage):
        """
        Parse Package homepage
        - pkg_homepage: Python str/unicode
        """
        if isinstance(pkg_homepage, str):
            try:
                return self.builder.set_pkg_home(self.document, pkg_homepage)
            except SPDXValueError:
                self.value_error("PKG_HOMEPAGE", pkg_homepage)
            except CardinalityError:
                self.more_than_one_error("PKG_HOMEPAGE")
            except OrderError:
                self.order_error("PKG_HOMEPAGE", "PKG_NAME")
        elif pkg_homepage is not None:
            self.value_error("PKG_HOMEPAGE", pkg_homepage)

    def parse_pkg_source_info(self, pkg_source_info):
        """
        Parse Package source information
        - pkg_source_info: Python str/unicode
        """
        if isinstance(pkg_source_info, str):
            try:
                return self.builder.set_pkg_source_info(self.document, pkg_source_info)
            except CardinalityError:
                self.more_than_one_error("PKG_SRC_INFO")
            except OrderError:
                self.order_error("PKG_SRC_INFO", "PKG_NAME")
        elif pkg_source_info is not None:
            self.value_error("PKG_SRC_INFO", pkg_source_info)

    def parse_pkg_concluded_license(self, pkg_concluded_license):
        """
        Parse Package concluded license
        - pkg_concluded_license: Python str/unicode
        """
        if isinstance(pkg_concluded_license, str):
            lic_parser = utils.LicenseListParser()
            lic_parser.build(write_tables=0, debug=0)
            license_object = self.replace_license(
                lic_parser.parse(pkg_concluded_license)
            )
            try:
                return self.builder.set_pkg_licenses_concluded(
                    self.document, license_object
                )
            except SPDXValueError:
                self.value_error("PKG_SINGLE_LICS", pkg_concluded_license)
            except CardinalityError:
                self.more_than_one_error("PKG_SINGLE_LICS")
            except OrderError:
                self.order_error("PKG_SINGLE_LICS", "PKG_NAME")
        elif pkg_concluded_license is not None:
            self.value_error("PKG_SINGLE_LICS", pkg_concluded_license)

    def parse_pkg_license_info_from_files(self, license_info_from_files):
        """
        Parse Package license information from files
        - license_info_from_files: Python list of licenses information from files (str/unicode)
        """
        if not self.package.are_files_analyzed:
            if not license_info_from_files:
                return
            self.value_error("PKG_LIC_FRM_FILES", license_info_from_files)
        if isinstance(license_info_from_files, list):
            for license_info_from_file in license_info_from_files:
                if isinstance(license_info_from_file, str):
                    lic_parser = utils.LicenseListParser()
                    lic_parser.build(write_tables=0, debug=0)
                    license_object = self.replace_license(
                        lic_parser.parse(license_info_from_file)
                    )
                    try:
                        self.builder.set_pkg_license_from_file(
                            self.document, license_object
                        )
                    except SPDXValueError:
                        self.value_error("PKG_LIC_FRM_FILES", license_info_from_file)
                    except OrderError:
                        self.order_error("PKG_LIC_FRM_FILES", "PKG_NAME")
                else:
                    self.value_error("PKG_LIC_FRM_FILES", license_info_from_file)
        elif license_info_from_files is not None:
            self.value_error("PKG_LIC_FRM_FILES_FIELD", license_info_from_files)

    def parse_pkg_attribution_text(self, pkg_attribution_texts):
        """
        Parse Package attribution texts
        - pkg_attribution_texts: list in yaml, json and string in xml format
        """
        if isinstance(pkg_attribution_texts, list) or isinstance(
            pkg_attribution_texts, str
        ):
            for pkg_attribution_text in pkg_attribution_texts:
                try:
                    return self.builder.set_pkg_attribution_text(
                        self.document, pkg_attribution_text
                    )
                except CardinalityError:
                    self.more_than_one_error("PKG_ATTRIBUTION_TEXT")
                except OrderError:
                    self.order_error("PKG_ATTRIBUTION_TEXT", "PKG_NAME")
            else:
                self.value_error("PKG_ATTRIBUTION_TEXT", pkg_attribution_texts)

    def parse_pkg_declared_license(self, pkg_declared_license):
        """
        Parse Package license declared
        - pkg_declared_license: Python str/unicode
        """
        if isinstance(pkg_declared_license, str):
            lic_parser = utils.LicenseListParser()
            lic_parser.build(write_tables=0, debug=0)
            license_object = self.replace_license(
                lic_parser.parse(pkg_declared_license)
            )
            try:
                return self.builder.set_pkg_license_declared(
                    self.document, license_object
                )
            except SPDXValueError:
                self.value_error("PKG_DECL_LIC", pkg_declared_license)
            except CardinalityError:
                self.more_than_one_error("PKG_DECL_LIC")
            except OrderError:
                self.order_error("PKG_DECL_LIC", "PKG_NAME")
        elif pkg_declared_license is not None:
            self.value_error("PKG_DECL_LIC", pkg_declared_license)

    def parse_pkg_license_comment(self, pkg_license_comment):
        """
        Parse Package license comment
        - pkg_license_comment: Python str/unicode
        """
        if isinstance(pkg_license_comment, str):
            try:
                return self.builder.set_pkg_license_comment(
                    self.document, pkg_license_comment
                )
            except CardinalityError:
                self.more_than_one_error("PKG_LIC_COMMENT")
            except OrderError:
                self.order_error("PKG_LIC_COMMENT", "PKG_NAME")
        elif pkg_license_comment is not None:
            self.value_error("PKG_LIC_COMMENT", pkg_license_comment)

    def parse_pkg_copyright_text(self, pkg_copyright_text):
        """
        Parse Package copyright text
        - pkg_copyright_text: Python str/unicode
        """
        if isinstance(pkg_copyright_text, str):
            try:
                return self.builder.set_pkg_cr_text(self.document, pkg_copyright_text)
            except CardinalityError:
                self.more_than_one_error("PKG_COPYRIGHT_TEXT")
            except OrderError:
                self.order_error("PKG_COPYRIGHT_TEXT", "PKG_NAME")
        elif pkg_copyright_text is not None:
            self.value_error("PKG_COPYRIGHT_TEXT", pkg_copyright_text)

    def parse_pkg_summary(self, pkg_summary):
        """
        Parse Package summary
        - pkg_summary: Python str/unicode
        """
        if isinstance(pkg_summary, str):
            try:
                return self.builder.set_pkg_summary(self.document, pkg_summary)
            except CardinalityError:
                self.more_than_one_error("PKG_SUMMARY")
            except OrderError:
                self.order_error("PKG_SUMMARY", "PKG_NAME")
        elif pkg_summary is not None:
            self.value_error("PKG_SUMMARY", pkg_summary)

    def parse_pkg_comment(self, pkg_comment):
        """
        Parse Package comment
        - pkg_comment: Python str/unicode
        """
        if isinstance(pkg_comment, str):
            try:
                return self.builder.set_pkg_comment(self.document, pkg_comment)
            except CardinalityError:
                self.more_than_one_error("PKG_COMMENT")
            except OrderError:
                self.order_error("PKG_COMMENT", "PKG_NAME")
        elif pkg_comment is not None:
            self.value_error("PKG_COMMENT", pkg_comment)

    def parse_pkg_description(self, pkg_description):
        """
        Parse Package description
        - pkg_description: Python str/unicode
        """
        if isinstance(pkg_description, str):
            try:
                return self.builder.set_pkg_desc(self.document, pkg_description)
            except CardinalityError:
                self.more_than_one_error("PKG_DESCRIPTION")
            except OrderError:
                self.order_error("PKG_DESCRIPTION", "PKG_NAME")
        elif pkg_description is not None:
            self.value_error("PKG_DESCRIPTION", pkg_description)

    def parse_pkg_files(self, pkg_has_files: List[str], method_to_parse_relationship: Callable) -> None:
        """
        Parse Package files
        - pkg_has_files: Python list of spdx_ids
        """
        if not self.package.are_files_analyzed:
            if not pkg_has_files:
                return
            self.value_error("PKG_FILES", pkg_has_files)

        if isinstance(pkg_has_files, list):
            for pkg_file_spdx_id in pkg_has_files:
                if isinstance(pkg_file_spdx_id, str):
                    method_to_parse_relationship(self.package.spdx_id, "CONTAINS", pkg_file_spdx_id)
                else:
                    self.value_error("PKG_FILE", pkg_file_spdx_id)
        elif pkg_has_files is not None:
            self.value_error("PKG_HAS_FILES", pkg_has_files)

    def parse_pkg_checksums(self, pkg_checksums: List[Dict]) -> Optional[bool]:
        """
        Parse Package checksums
        - pkg_chksums: Python List
        """
        if isinstance(pkg_checksums, list):
            for checksum in pkg_checksums:
                self.builder.set_pkg_checksum(self.document, checksum)
            return True
        if isinstance(pkg_checksums, str):
            # kept for backwards compatibility
            try:
                return self.builder.set_pkg_checksum(self.document, pkg_checksums)
            except CardinalityError:
                self.more_than_one_error("PKG_CHECKSUM")
            except OrderError:
                self.order_error("PKG_CHECKSUM", "PKG_NAME")
        elif pkg_checksums is not None:
            self.value_error("PKG_CHECKSUM", pkg_checksums)

    def parse_package_external_refs(self, external_refs: List[Dict]):
        if external_refs is None:
            return
        if not isinstance(external_refs, list):
            self.value_error("PACKAGE_EXTERNAL_REFS", external_refs)
            return

        for external_ref_dict in external_refs:
            external_ref = ExternalPackageRef(category=external_ref_dict["referenceCategory"],
                                              pkg_ext_ref_type=external_ref_dict["referenceType"],
                                              locator=external_ref_dict["referenceLocator"])
            if "comment" in external_ref_dict:
                external_ref.comment = external_ref_dict["comment"]
            self.package.add_pkg_ext_refs(external_ref)

    def parse_primary_package_purpose(self, primary_package_purpose: str):
        if primary_package_purpose is None:
            return
        primary_package_purpose = primary_package_purpose.replace("-", "_")  # OPERATING-SYSTEM -> OPERATING_SYSTEM
        if primary_package_purpose not in [purpose.name for purpose in PackagePurpose]:
            self.value_error("PRIMARY_PACKAGE_PURPOSE", primary_package_purpose)
            return

        purpose_enum = PackagePurpose[primary_package_purpose]
        self.package.primary_package_purpose = purpose_enum

    def parse_release_date(self, release_date: str):
        if release_date is None:
            return

        parsed_date: datetime = utils.datetime_from_iso_format(release_date)
        if parsed_date is not None:
            self.package.release_date = parsed_date
        else:
            self.value_error("RELEASE_DATE", release_date)

    def parse_built_date(self, built_date: str):
        if built_date is None:
            return

        parsed_date: datetime = utils.datetime_from_iso_format(built_date)
        if parsed_date is not None:
            self.package.built_date = parsed_date
        else:
            self.value_error("BUILT_DATE", built_date)

    def parse_valid_until_date(self, valid_until_date: str):
        if valid_until_date is None:
            return

        parsed_date: datetime = utils.datetime_from_iso_format(valid_until_date)
        if parsed_date is not None:
            self.package.valid_until_date = parsed_date
        else:
            self.value_error("VALID_UNTIL_DATE", valid_until_date)

def flatten_document(document):
    """
    Flatten document to match current data model. File objects are nested within packages according to hasFiles-tag.
    """
    files_by_id = {}
    if "files" in document:
        for f in document.get("files"):
            for checksum in f["checksums"]:
                if checksum["algorithm"] == "SHA1" or "sha1" in checksum["algorithm"]:
                    f["sha1"] = checksum["checksumValue"]
                    break
            files_by_id[f["SPDXID"]] = f
    if "packages" in document:
        packages = document.get("packages")
        for package in packages:
            if "hasFiles" in package:
                package["files"] = [{
                    "File": files_by_id[spdxid.split("#")[-1]]} for spdxid in package["hasFiles"]
                ]
            for checksum in package.get("checksums", []):
                if checksum["algorithm"] == "SHA1" or "sha1" in checksum["algorithm"]:
                    package["sha1"] = checksum["checksumValue"]
                    break

    return document


class Parser(
    CreationInfoParser,
    ExternalDocumentRefsParser,
    LicenseParser,
    AnnotationParser,
    RelationshipParser,
    SnippetParser,
    ReviewParser,
    FileParser,
    PackageParser,
):
    def __init__(self, builder, logger):
        super(Parser, self).__init__(builder, logger)

    def json_yaml_set_document(self, data):
        # we could verify that the spdxVersion >= 2.2, but we try to be resilient in parsing
        if data.get("spdxVersion"):
            self.document_object = data
            return
        self.document_object = data.get("Document")

    def parse(self):
        """
        Parse Document Information fields
        """
        self.error = False
        self.document = document.Document()
        self.document_object = flatten_document(self.document_object)
        if not isinstance(self.document_object, dict):
            self.logger.log("Empty or not valid SPDX Document")
            self.error = True
            return self.document, self.error

        self.parse_doc_version(self.document_object.get("spdxVersion"))
        self.parse_doc_data_license(self.document_object.get("dataLicense"))
        self.parse_doc_id(self.document_object.get("SPDXID"))
        self.parse_doc_name(self.document_object.get("name"))
        self.parse_doc_namespace(self.document_object.get("documentNamespace"))
        self.parse_doc_comment(self.document_object.get("comment"))
        self.parse_creation_info(self.document_object.get("creationInfo"))
        self.parse_external_document_refs(
            self.document_object.get("externalDocumentRefs")
        )
        self.parse_extracted_license_info(
            self.document_object.get("hasExtractedLicensingInfos")
        )
        self.parse_annotations(self.document_object.get("annotations"), spdx_id=self.document_object.get("SPDXID"))
        self.parse_relationships(self.document_object.get("relationships"))
        self.parse_reviews(self.document_object.get("reviewers"))
        self.parse_snippets(self.document_object.get("snippets"))

        self.parse_packages(self.document_object.get("packages"))
        self.parse_files(self.document_object.get("files"))

        if self.document_object.get("documentDescribes"):
            self.parse_doc_described_objects(self.document_object.get("documentDescribes"))

        validation_messages = ErrorMessages()
        # Report extra errors if self.error is False otherwise there will be
        # redundant messages
        validation_messages = self.document.validate(validation_messages)
        if not self.error:
            if validation_messages:
                for msg in validation_messages:
                    self.logger.log(msg)
                self.error = True

        return self.document, self.error

    def parse_doc_version(self, doc_version):
        """
        Parse Document version
        - doc_version: Python str/unicode
        """
        if isinstance(doc_version, str):
            try:
                return self.builder.set_doc_version(self.document, doc_version)
            except SPDXValueError:
                self.value_error("DOC_VERS_VALUE", doc_version)
            except CardinalityError:
                self.more_than_one_error("DOC_VERS_VALUE")
        else:
            self.value_error("DOC_VERS_VALUE", doc_version)

    def parse_doc_data_license(self, doc_data_license):
        """
        Parse Document data license
        - doc_data_license: Python str/unicode
        """
        try:
            return self.builder.set_doc_data_lics(self.document, doc_data_license)
        except SPDXValueError:
            self.value_error("DOC_D_LICS", doc_data_license)
        except CardinalityError:
            self.more_than_one_error("DOC_D_LICS")

    def parse_doc_id(self, doc_id):
        """
        Parse Document SPDX id
        - doc_id: Python str/unicode
        """
        if isinstance(doc_id, str):
            try:
                return self.builder.set_doc_spdx_id(self.document, doc_id)
            except SPDXValueError:
                self.value_error("DOC_SPDX_ID_VALUE", doc_id)
            except CardinalityError:
                self.more_than_one_error("DOC_SPDX_ID_VALUE")
        else:
            self.value_error("DOC_SPDX_ID_VALUE", doc_id)

    def parse_doc_name(self, doc_name):
        """
        Parse Document name
        - doc_name: Python str/unicode
        """
        if isinstance(doc_name, str):
            try:
                return self.builder.set_doc_name(self.document, doc_name)
            except CardinalityError:
                self.more_than_one_error("DOC_NAME_VALUE")
        else:
            self.value_error("DOC_NAME_VALUE", doc_name)

    def parse_doc_namespace(self, doc_namespace):
        """
        Parse Document namespace
        - doc_namespace: Python str/unicode
        """
        if isinstance(doc_namespace, str):
            try:
                return self.builder.set_doc_namespace(self.document, doc_namespace)
            except SPDXValueError:
                self.value_error("DOC_NAMESPACE_VALUE", doc_namespace)
            except CardinalityError:
                self.more_than_one_error("DOC_NAMESPACE_VALUE")
        else:
            self.value_error("DOC_NAMESPACE_VALUE", doc_namespace)

    def parse_doc_comment(self, doc_comment):
        """
        Parse Document comment
        - doc_comment: Python str/unicode
        """
        if isinstance(doc_comment, str):
            try:
                return self.builder.set_doc_comment(self.document, doc_comment)
            except CardinalityError:
                self.more_than_one_error("DOC_COMMENT_VALUE")
        elif doc_comment is not None:
            self.value_error("DOC_COMMENT_VALUE", doc_comment)

    def parse_doc_described_objects(self, doc_described_objects):
        """
        Parse Document documentDescribes (SPDXIDs)
        - doc_described_objects: Python list of strings
        """
        if isinstance(doc_described_objects, list):
            described_spdxids = filter(
                lambda described: isinstance(described, str), doc_described_objects
            )
            for spdxid in described_spdxids:
                self.parse_relationship(self.document.spdx_id, "DESCRIBES", spdxid)

            return True
        else:
            self.value_error("DOC_DESCRIBES", doc_described_objects)

    def parse_packages(self, packages):
        """
        Parse SPDXLite packages list
        """
        if packages is None:
            return
        if isinstance(packages, list):
            for package in packages:
                self.parse_package(package, self.parse_relationship)
            return True
        else:
            self.value_error("PACKAGES", packages)



# Copyright (c) 2014 Ahmed H. Ismail
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import total_ordering
import re


@total_ordering
class Version(object):
    """
    Version number composed of major and minor.
    Fields:
    - major: Major number, int.
    - minor: Minor number, int.
    """

    VERS_STR_REGEX = re.compile(r"(\d+)\.(\d+)")

    def __init__(self, major, minor):
        self.major = int(major)
        self.minor = int(minor)

    @classmethod
    def from_str(cls, value):
        """Constructs a Version from a string.
        Returns None if string not in N.N form where N represents a
        number.
        """
        m = cls.VERS_STR_REGEX.match(value)
        if m is not None:
            return cls(int(m.group(1)), int(m.group(2)))
        else:
            return None

    def __repr__(self):
        return "Version" + repr((self.major, self.minor))

    def __str__(self):
        return "SPDX-{major}.{minor}".format(**self.__dict__)

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and self.major == other.major
            and self.minor == other.minor
        )

    def __lt__(self, other):
        return self.major < other.major or (
            self.major == other.major and self.minor < other.minor
        )



# Copyright (c) 2014 Ahmed H. Ismail
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from spdx.file import File
from spdx.license import ExtractedLicense
from spdx.parsers.loggers import ErrorMessages

import warnings

from functools import total_ordering
from spdx.relationship import Relationship

from spdx.relationship import Relationship


@total_ordering
class ExternalDocumentRef(object):
    """
    External Document References entity that contains the following fields :
    - external_document_id: A unique string containing letters, numbers, '.',
        '-' or '+'.
    - spdx_document_uri: The unique ID of the SPDX document being referenced.
    - check_sum: The checksum of the referenced SPDX document.
    """

    def __init__(
        self, external_document_id=None, spdx_document_uri=None, checksum=None
    ):
        self.external_document_id = external_document_id
        self.spdx_document_uri = spdx_document_uri
        self.checksum = checksum

    def __eq__(self, other):
        return (
            isinstance(other, ExternalDocumentRef)
            and self.external_document_id == other.external_document_id
            and self.spdx_document_uri == other.spdx_document_uri
            and self.checksum == other.checksum
        )

    def __lt__(self, other):
        return (self.external_document_id, self.spdx_document_uri, self.checksum) < (
            other.external_document_id,
            other.spdx_document_uri,
            other.checksum,
        )

    def validate(self, messages: ErrorMessages) -> ErrorMessages:
        """
        Check that all the fields are valid.
        Appends any error messages to messages parameter shall be a ErrorMessages.
        """
        self.validate_ext_doc_id(messages)
        self.validate_spdx_doc_uri(messages)
        self.validate_checksum(messages)
        return messages

    def validate_ext_doc_id(self, messages):
        if not self.external_document_id:
            messages.append("ExternalDocumentRef has no External Document ID.")

    def validate_spdx_doc_uri(self, messages):
        if not self.spdx_document_uri:
            messages.append("ExternalDocumentRef has no SPDX Document URI.")

    def validate_checksum(self, messages):
        if not self.checksum:
            messages.append("ExternalDocumentRef has no Checksum.")


class Document(object):
    """
    Represent an SPDX document with these fields:
    - version: Spec version. Mandatory, one - Type: Version.
    - data_license: SPDX-Metadata license. Mandatory, one. Type: License.
    - name: Name of the document. Mandatory, one. Type: str.
    - spdx_id: SPDX Identifier for the document to refer to itself in
      relationship to other elements. Mandatory, one. Type: str.
    - ext_document_references: External SPDX documents referenced within the
        given SPDX document. Optional, one or many. Type: ExternalDocumentRef
    - comment: Comments on the SPDX file, optional one. Type: str
    - documentNamespace: SPDX document specific namespace. Mandatory, one. Type: str
    - creation_info: SPDX file creation info. Mandatory, one. Type: CreationInfo
    - package: Package described by this document. Mandatory, one. Type: Package
    - extracted_licenses: List of licenses extracted that are not part of the
      SPDX license list. Optional, many. Type: ExtractedLicense.
    - reviews: SPDX document review information, Optional zero or more.
      Type: Review.
    - annotations: SPDX document annotation information, Optional zero or more.
      Type: Annotation.
    - snippet: Snippet information. Optional zero or more. Type: Snippet.
    - relationships: Relationship between two SPDX elements. Optional zero or more.
      Type: Relationship. 
    """

    def __init__(
        self,
        version=None,
        data_license=None,
        name=None,
        spdx_id=None,
        namespace=None,
        comment=None,
        package=None,
        license_list_version=None,
    ):
        # avoid recursive import
        from spdx.creationinfo import CreationInfo

        self.version = version
        self.data_license = data_license
        self.name = name
        self.spdx_id = spdx_id
        self.ext_document_references = []
        self.comment = comment
        self.namespace = namespace
        self.creation_info = CreationInfo()
        self.files: List['File'] = []
        self.packages = []
        if package is not None:
            self.packages.append(package)
        self.extracted_licenses = []
        self.reviews = []
        self.annotations = []
        self.relationships: List[Relationship] = []
        self.snippet = []

        # due to backwards compatibility write input argument for license list version to creation info
        if license_list_version:
            self.creation_info.set_license_list_version(license_list_version)

    def add_review(self, review):
        self.reviews.append(review)

    def add_annotation(self, annotation):
        self.annotations.append(annotation)

    def add_relationship(self, relationship):
        self.relationships.append(relationship)

    def add_extr_lic(self, lic):
        self.extracted_licenses.append(lic)

    def add_ext_document_reference(self, ext_doc_ref):
        self.ext_document_references.append(ext_doc_ref)

    def add_snippet(self, snip):
        self.snippet.append(snip)

    def add_package(self, package):
        self.packages.append(package)

    def add_file(self, file: 'File') -> None:
        self.files.append(file)

    # For backwards compatibility with older versions, we support a
    # mode where the first package in a document may be referred to as
    # the document's "package", and the files it contains may be
    # referred to as the document's files.  This usage is deprecated.

    @property
    def package(self):
        warnings.warn('document.package and document.files are deprecated; '
                      'use document.packages instead',
                      DeprecationWarning)
        if len(self.packages) == 0:
            return None
        else:
            return self.packages[0]

    @package.setter
    def package(self, value):
        warnings.warn('document.package and document.files are deprecated; '
                      'use document.packages instead',
                      DeprecationWarning)
        if len(self.packages) == 0:
            self.packages.append(value)
        else:
            self.packages[0] = value

    @property
    def has_comment(self):
        return self.comment is not None

    def validate(self, messages=None):
        """
        Validate all fields of the document and update the
        messages list with user friendly error messages for display.
        """
        if isinstance(messages, list):
            raise TypeError("messages should be None or an instance of ErrorMessages")
        if messages is None:
            messages = ErrorMessages()

        messages.push_context(self.name)
        self.validate_version(messages)
        self.validate_data_lics(messages)
        self.validate_name(messages)
        self.validate_spdx_id(messages)
        self.validate_namespace(messages)
        self.validate_ext_document_references(messages)
        self.validate_creation_info(messages)
        self.validate_files(messages)
        self.validate_packages(messages)
        self.validate_extracted_licenses(messages)
        self.validate_reviews(messages)
        self.validate_snippet(messages)
        self.validate_annotations(messages)
        self.validate_relationships(messages)
        messages.pop_context()
        return messages

    def validate_version(self, messages):
        if self.version is None:
            messages.append("Document has no version.")

    def validate_data_lics(self, messages):
        if self.data_license is None:
            messages.append("Document has no data license.")
        else:
            # FIXME: REALLY? what if someone wants to use something else?
            if self.data_license.identifier != "CC0-1.0":
                messages.append("Document data license must be CC0-1.0.")

    def validate_name(self, messages):
        if self.name is None:
            messages.append("Document has no name.")

    def validate_namespace(self, messages):
        if self.namespace is None:
            messages.append("Document has no namespace.")

    def validate_spdx_id(self, messages):
        if self.spdx_id is None:
            messages.append("Document has no SPDX Identifier.")
        else:
            if not self.spdx_id.endswith("SPDXRef-DOCUMENT"):
                messages.append("Invalid Document SPDX Identifier value.")

    def validate_ext_document_references(self, messages):
        for doc in self.ext_document_references:
            if isinstance(doc, ExternalDocumentRef):
                messages = doc.validate(messages)
            else:
                messages = list(messages) + [
                    "External document references must be of the type "
                    "spdx.document.ExternalDocumentRef and not " + str(type(doc))
                ]

    def validate_reviews(self, messages):
        for review in self.reviews:
            messages = review.validate(messages)

    def validate_files(self, messages: ErrorMessages) -> None:
        for file in self.files:
            messages = file.validate(messages)

    def validate_annotations(self, messages):
        for annotation in self.annotations:
            messages = annotation.validate(messages)

    def validate_relationships(self, messages):
        for relationship in self.relationships:
            messages = relationship.validate(messages)

    def validate_snippet(self, messages=None):
        for snippet in self.snippet:
            snippet.validate(messages)

    def validate_creation_info(self, messages):
        if self.creation_info is not None:
            self.creation_info.validate(messages)
        else:
            messages.append("Document has no creation information.")

    def validate_packages(self, messages):
        for package in self.packages:
            messages = package.validate(messages)

    def validate_extracted_licenses(self, messages):
        for lic in self.extracted_licenses:
            if isinstance(lic, ExtractedLicense):
                messages = lic.validate(messages)
            else:
                messages.append(
                    "Document extracted licenses must be of type "
                    "spdx.document.ExtractedLicense and not " + type(lic)
                )



# Copyright (c) 2014 Ahmed H. Ismail
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class StandardLogger(object):
    def log(self, msg):
        print(msg)


class FileLogger(object):
    def __init__(self, logfile):
        self.dest = logfile

    def log(self, msg):
        self.dest.write(msg + "\n")


class ErrorMessages:

    def __init__(self):
        self.messages = []
        self.context = []
    
    def push_context(self, context):
        """push some context information to better identify where is the problem"""
        self.context.append(context)

    def pop_context(self):
        """pop the last context information"""
        self.context.pop()
    
    def append(self, message, *args, **kwargs):
        """add a message with standard python format
        the current context is prefixed to the message
        """
        message = message.format(*args, **kwargs)
        message = "".join([c + ": " for c in self.context if c]) + message
        self.messages.append(message)

    def __iter__(self):
        return self.messages.__iter__()

    def __bool__(self):
        return len(self.messages)>0

    def __nonzero__(self):
        return len(self.messages)>0
    
    def __eq__(self, b):
        if isinstance(b, ErrorMessages):
            return self.messages == b.messages
        return self.messages == b



#!/usr/bin/env python
from spdx.relationship import Relationship

# Writes a new tag/value file from scratch.
# Usage: write_tv <tagvaluefile>
if __name__ == "__main__":
    import sys
    import codecs
    from spdx.writers.tagvalue import write_document, InvalidDocumentError
    from spdx.parsers.loggers import ErrorMessages
    from spdx.document import Document
    from spdx.license import License, LicenseConjunction, ExtractedLicense
    from spdx.version import Version
    from spdx.creationinfo import Person
    from spdx.review import Review
    from spdx.package import Package
    from spdx.file import File, FileType
    from spdx.checksum import Checksum
    from spdx.utils import SPDXNone, NoAssert, UnKnown

    doc = Document()
    doc.version = Version(1, 2)
    doc.name = "Hello_SPDX"
    doc.spdx_id = "Test#SPDXRef-DOCUMENT"
    doc.comment = "Example Document"
    doc.namespace = "https://spdx.org/spdxdocs/spdx-example-444504E0-4F89-41D3-9A0C-0305E82C3301"
    doc.data_license = License.from_identifier("CC0-1.0")
    doc.creation_info.add_creator(Person("Alice", "alice@example.com"))
    doc.creation_info.set_created_now()
    review = Review(Person("Joe", None))
    review.set_review_date_now()
    review.comment = "Joe reviewed this document"
    doc.add_review(review)
    # File
    testfile1 = File("TestFile1")
    testfile1.type = FileType.BINARY
    testfile1.spdx_id = "TestFilet#SPDXRef-FILE"
    testfile1.comment = "This is a test file."
    testfile1.chksum = Checksum("SHA1", "c537c5d99eca5333f23491d47ededd083fefb7ad")
    testfile1.conc_lics = License.from_identifier("BSD-2-Clause")
    testfile1.add_lics(License.from_identifier("BSD-2-Clause"))
    testfile1.copyright = SPDXNone()
    testfile1.add_artifact("name", "TagWriteTest")
    testfile1.add_artifact("home", UnKnown())
    testfile1.add_artifact("uri", "http://tagwritetest.test")

    testfile2 = File("TestFile2")
    testfile2.type = FileType.SOURCE
    testfile2.spdx_id = "TestFile2#SPDXRef-FILE"
    testfile2.comment = "This is a test file."
    testfile2.chksum = Checksum("SHA1", "bb154f28d1cf0646ae21bb0bec6c669a2b90e113")
    testfile2.conc_lics = License.from_identifier("Apache-2.0")
    testfile2.add_lics(License.from_identifier("Apache-2.0"))
    testfile2.copyright = NoAssert()
    doc.add_file(testfile1)
    doc.add_file(testfile2)

    # Package
    package = Package()
    package.name = "TagWriteTest"
    package.version = "1.0"
    package.file_name = "twt.jar"
    package.spdx_id = 'TestPackage#SPDXRef-PACKAGE'
    package.download_location = "http://www.tagwritetest.test/download"
    package.checksum = Checksum("SHA1", "c537c5d99eca5333f23491d47ededd083fefb7ad")
    package.homepage = SPDXNone()
    package.verif_code = "4e3211c67a2d28fced849ee1bb76e7391b93feba"
    license_set = LicenseConjunction(
        License.from_identifier("Apache-2.0"), License.from_identifier("BSD-2-Clause")
    )
    package.conc_lics = license_set
    package.license_declared = license_set
    package.add_lics_from_file(License.from_identifier("Apache-2.0"))
    package.add_lics_from_file(License.from_identifier("BSD-2-Clause"))
    package.cr_text = NoAssert()
    package.summary = "Simple package."
    package.description = "Really simple package."

    doc.package = package
    relationship = Relationship("TestPackage#SPDXRef-PACKAGE CONTAINS TestFilet#SPDXRef-FILE")
    doc.add_relationship(relationship)
    relationship = Relationship("TestPackage#SPDXRef-PACKAGE CONTAINS TestFile2#SPDXRef-FILE")
    doc.add_relationship(relationship)

    # An extracted license

    lic = ExtractedLicense("LicenseRef-1")
    lic.text = "Some non legal legal text.."
    doc.add_extr_lic(lic)

    file = sys.argv[1]
    with codecs.open(file, mode="w", encoding="utf-8") as out:
        try:
            write_document(doc, out)
        except InvalidDocumentError as e:
            print("Document is Invalid:\n\t", end="")
            print("\n\t".join(e.args[0]))
            messages = ErrorMessages()
            doc.validate(messages)
            print("\n".join(messages.messages))



#!/usr/bin/env python

# Copyright (c) 2020 Yash Varshney
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#    http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from spdx import utils
from spdx.parsers.parse_anything import parse_file
import spdx.file as spdxfile

import click


@click.command()
@click.option("--file", prompt="File name", help="The file to be parsed")
@click.option("--force", is_flag=True, help="print information even if there are some parsing errors")
def main(file, force):
    """
    COMMAND-LINE TOOL for parsing file of RDF, XML, JSON, YAML and XML format.

    To use : run `pyspdxtools_parser` using terminal or run `pyspdxtools_parser --file <file name>`

    """
    doc, errors = parse_file(file)
    if errors:
        print("Errors while parsing: ", errors)
        if not force:
            return 1

    print("doc comment: {0}".format(doc.comment))
    print("Creators:")
    for c in doc.creation_info.creators:
        print("\t{0}".format(c))
    print("Document review information:")
    for review in doc.reviews:
        print("\tReviewer: {0}".format(review.reviewer))
        print("\tDate: {0}".format(review.review_date))
        print("\tComment: {0}".format(review.comment))
    print("Creation comment: {0}".format(doc.creation_info.comment))
    for package in doc.packages:
        print("Package Name: {0}".format(package.name))
        print("Package Version: {0}".format(package.version))
        print(
            "Package Download Location: {0}".format(package.download_location)
        )
        print("Package Homepage: {0}".format(package.homepage))
        for checksum in doc.package.checksums.values():
            print("Package Checksum: {0} {1}".format(checksum.identifier.name, checksum.value))
        print("Package Attribution Text: {0}".format(package.attribution_text))
        print("Package verification code: {0}".format(package.verif_code))
        print(
            "Package excluded from verif: {0}".format(
                ",".join(package.verif_exc_files)
            )
        )
        print("Package license concluded: {0}".format(package.conc_lics))
        print("Package license declared: {0}".format(package.license_declared))
        print("Package licenses from files:")
        for lics in package.licenses_from_files:
            print("\t{0}".format(lics))
        print("Package Copyright text: {0}".format(package.cr_text))
        print("Package summary: {0}".format(package.summary))
        print("Package description: {0}".format(package.description))
        if len(package.pkg_ext_refs) > 0:
            print("Package external references:")
            for ref in package.pkg_ext_refs:
                print(f"\tCategory: {ref.category}")
                print(f"\tType: {ref.pkg_ext_ref_type}")
                print(f"\tLocator: {ref.locator}")
                if ref.comment:
                    print(f"\tComment: {ref.comment}")
    if doc.files:
        print("Files:")
    for f in doc.files:
        print("\tFile name: {0}".format(f.name))
        for file_type in f.file_types:
            print("\tFile type: {0}".format(file_type.name))
        for file_checksum in f.checksums.values():
            print("\tFile Checksum: {0} {1}".format(file_checksum.identifier.name, file_checksum.value))
        print("\tFile license concluded: {0}".format(f.conc_lics))
        print(
            "\tFile license info in file: {0}".format(
                ",".join(
                    map(lambda l: l.identifier if not isinstance(l, (utils.SPDXNone, utils.NoAssert)) else l.to_value(),
                        f.licenses_in_file))
            )
        )
        print(
            "\tFile artifact of project name: {0}".format(
                ",".join(f.artifact_of_project_name)
            )
        )

    if doc.extracted_licenses:
        print("Document Extracted licenses:")
    for lics in doc.extracted_licenses:
        print("\tIdentifier: {0}".format(lics.identifier))
        print("\tName: {0}".format(lics.full_name))
        print("\License Text: {0}".format(lics.text))
    if doc.annotations:
        print("Annotations:")
    for an in doc.annotations:
        print("\tAnnotator: {0}".format(an.annotator))
        print("\tAnnotation Date: {0}".format(an.annotation_date))
        print("\tAnnotation Comment: {0}".format(an.comment))
        print("\tAnnotation Type: {0}".format(an.annotation_type))
        print("\tAnnotation SPDX Identifier: {0}".format(an.spdx_id))

    if doc.relationships:
        print("Relationships: ")
    for relation in doc.relationships:
        print("\tRelationship: {0}".format(relation.relationship))
        try:
            print("\tRelationship: {0}".format(relation.comment))
        except:
            continue


if __name__ == "__main__":
    main()



# Copyright (c) 2014 Ahmed H. Ismail
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from functools import reduce

from rdflib import Graph
from rdflib import Namespace
from rdflib import RDF
from rdflib import RDFS

from spdx import document
from spdx import license
from spdx import utils
from spdx.checksum import Checksum, ChecksumAlgorithm
from spdx.parsers.builderexceptions import CardinalityError
from spdx.parsers.builderexceptions import SPDXValueError
from spdx.parsers.loggers import ErrorMessages


ERROR_MESSAGES = {
    "DOC_VERS_VALUE": "Invalid specVersion '{0}' must be SPDX-M.N where M and N are numbers.",
    "DOC_D_LICS": "Invalid dataLicense '{0}' must be http://spdx.org/licenses/CC0-1.0.",
    "DOC_SPDX_ID_VALUE": "Invalid SPDXID value, SPDXID must be the document namespace appended "
    'by "#SPDXRef-DOCUMENT", line: {0}',
    "DOC_NAMESPACE_VALUE": 'Invalid DocumentNamespace value {0}, must contain a scheme (e.g. "https:") '
    'and should not contain the "#" delimiter.',
    "LL_VALUE": "Invalid licenseListVersion '{0}' must be of the format N.N where N is a number",
    "CREATED_VALUE": "Invalid created value '{0}' must be date in ISO 8601 format.",
    "CREATOR_VALUE": "Invalid creator value '{0}' must be Organization, Tool or Person.",
    "EXT_DOC_REF_VALUE": "Failed to extract {0} from ExternalDocumentRef.",
    "PKG_SPDX_ID_VALUE": 'SPDXID must be "SPDXRef-[idstring]" where [idstring] is a unique string containing '
    'letters, numbers, ".", "-".',
    "PKG_SUPPL_VALUE": "Invalid package supplier value '{0}' must be Organization, Person or NOASSERTION.",
    "PKG_ORIGINATOR_VALUE": "Invalid package supplier value '{0}'  must be Organization, Person or NOASSERTION.",
    "PKG_DOWN_LOC": "Invalid package download location value '{0}'  must be a url or NONE or NOASSERTION",
    "PKG_FILES_ANALYZED_VALUE": "FilesAnalyzed must be a boolean value, line: {0}",
    "PKG_CONC_LIST": "Package concluded license list must have more than one member",
    "LICS_LIST_MEMBER": "Declarative or Conjunctive license set member must be a license url or identifier",
    "PKG_SINGLE_LICS": "Package concluded license must be a license url or spdx:noassertion or spdx:none.",
    "PKG_LICS_INFO_FILES": "Package licenseInfoFromFiles must be a license or spdx:none or spdx:noassertion",
    "FILE_SPDX_ID_VALUE": 'SPDXID must be "SPDXRef-[idstring]" where [idstring] is a unique string containing '
    'letters, numbers, ".", "-".',
    "PKG_EXT_REF_CATEGORY": '\'{0}\' must be "SECURITY", "PACKAGE-MANAGER", or "OTHER".',
    "PKG_EXT_REF_TYPE": '{0} must be a unique string containing letters, numbers, ".", or "-".',
    "FILE_TYPE": "Unknown file type.",
    "FILE_SINGLE_LICS": "File concluded license must be a license url or spdx:noassertion or spdx:none.",
    "REVIEWER_VALUE": "Invalid reviewer value '{0}' must be Organization, Tool or Person.",
    "REVIEW_DATE": "Invalid review date value '{0}' must be date in ISO 8601 format.",
    "ANNOTATOR_VALUE": "Invalid annotator value '{0}' must be Organization, Tool or Person.",
    "ANNOTATION_DATE": "Invalid annotation date value '{0}' must be date in ISO 8601 format.",
    "SNIPPET_SPDX_ID_VALUE": 'SPDXID must be "SPDXRef-[idstring]" where [idstring] is a unique string '
    'containing letters, numbers, ".", "-".',
    "SNIPPET_SINGLE_LICS": "Snippet Concluded License must be a license url or spdx:noassertion or spdx:none.",
    "SNIPPET_LIC_INFO": "License Information in Snippet must be a license url or a reference "
    "to the license, denoted by LicenseRef-[idstring] or spdx:noassertion or spdx:none.",
    "RELATIONSHIP": "relationship type must be of supported type",
}


def convert_rdf_checksum_algorithm(rdf_checksum_algorithm: str) -> ChecksumAlgorithm:
    split_string = rdf_checksum_algorithm.split('#')
    if len(split_string) != 2:
        raise SPDXValueError('Unknown checksum algorithm {}'.format(rdf_checksum_algorithm))
    checksum_algorithm = ChecksumAlgorithm.checksum_from_rdf(split_string[1])
    return checksum_algorithm


class BaseParser(object):
    """
    Base class for all parsers.
    Contains logger, doap_namespace, spdx_namespace and model builder.
    Also provides utility functions used by the deriving parsers.
    """

    def __init__(self, builder, logger):
        self.logger = logger
        self.doap_namespace = Namespace("http://usefulinc.com/ns/doap#")
        self.spdx_namespace = Namespace("http://spdx.org/rdf/terms#")
        self.builder = builder

    def more_than_one_error(self, field):
        """
        Logs a more than one error.
        field is the field/property that has more than one defined.
        """
        msg = "More than one {0} defined.".format(field)
        self.logger.log(msg)
        self.error = True

    def value_error(self, key, bad_value):
        """
        Report a value error using ERROR_MESSAGES dict.
        key - key to use for ERROR_MESSAGES.
        bad_value - is passed to format which is called on what key maps to
        in ERROR_MESSAGES.
        """
        msg = ERROR_MESSAGES[key].format(bad_value)
        self.logger.log(msg)
        self.error = True

    def to_special_value(self, value):
        """
        Check if value is a special SPDX value such as
        NONE, NOASSERTION or UNKNOWN if so returns proper model.
        else returns value
        """
        if value == self.spdx_namespace.none or str(value) == "NONE":
            return utils.SPDXNone()
        elif value == self.spdx_namespace.noassertion or str(value) == "NOASSERTION":
            return utils.NoAssert()
        elif value == self.spdx_namespace.unknown or str(value) == "UNKNOWN":
            return utils.UnKnown()
        else:
            return str(value)


class LicenseParser(BaseParser):
    """
    Helper class for parsing extracted licenses and license lists.
    """

    LICS_REF_REGEX = re.compile("LicenseRef-.+", re.UNICODE)

    def __init__(self, builder, logger):
        super(LicenseParser, self).__init__(builder, logger)

    def handle_lics(self, lics):
        """
        Return a License from a `lics` license resource.
        """
        # Handle extracted licensing info type.
        if (
            lics,
            RDF.type,
            self.spdx_namespace["ExtractedLicensingInfo"],
        ) in self.graph:
            return self.parse_only_extr_license(lics)

        # Assume resource, hence the path separator
        ident_start = lics.rfind("/") + 1
        if ident_start == 0:
            # special values such as spdx:noassertion
            special = self.to_special_value(lics)
            if special == lics:
                if self.LICS_REF_REGEX.match(lics):
                    # Is a license ref i.e LicenseRef-1
                    return license.License.from_identifier(str(lics))
                else:
                    # Not a known license form
                    raise SPDXValueError("License")
            else:
                # is a special value
                return special
        else:
            # license url
            return license.License.from_identifier(lics[ident_start:])

    def get_extr_license_ident(self, extr_lic):
        """
        Return a license identifier from an ExtractedLicense or None.
        """
        identifier_triples = list(
            self.graph.triples((extr_lic, self.spdx_namespace["licenseId"], None))
        )

        if not identifier_triples:
            self.error = True
            msg = "Extracted license must have licenseId property."
            self.logger.log(msg)
            return

        if len(identifier_triples) > 1:
            self.more_than_one_error("extracted license identifier_triples")
            return

        identifier_triple = identifier_triples[0]
        _s, _p, identifier = identifier_triple
        return str(identifier)

    def get_extr_license_text(self, extr_lic):
        """
        Return extracted text from an ExtractedLicense or None.
        """
        text_triples = list(
            self.graph.triples((extr_lic, self.spdx_namespace["extractedText"], None))
        )
        if not text_triples:
            self.error = True
            msg = "Extracted license must have extractedText property"
            self.logger.log(msg)
            return

        if len(text_triples) > 1:
            self.more_than_one_error("extracted license text")
            return

        text_triple = text_triples[0]
        _s, _p, text = text_triple
        return str(text)

    def get_extr_lic_name(self, extr_lic):
        """
        Return the license name from an ExtractedLicense or None
        """
        extr_name_list = list(
            self.graph.triples((extr_lic, self.spdx_namespace["licenseName"], None))
        )
        if len(extr_name_list) > 1:
            self.more_than_one_error("extracted license name")
            return
        elif len(extr_name_list) == 0:
            return
        return str(self.to_special_value(extr_name_list[0][2]))

    def get_extr_lics_xref(self, extr_lic):
        """
        Return a list of cross references.
        """
        xrefs = list(self.graph.triples((extr_lic, RDFS.seeAlso, None)))
        return list(map(lambda xref_triple: xref_triple[2], xrefs))

    def get_extr_lics_comment(self, extr_lics):
        """
        Return license comment or None.
        """
        comment_list = list(self.graph.triples((extr_lics, RDFS.comment, None)))
        if len(comment_list) > 1:
            self.more_than_one_error("extracted license comment")
            return
        elif len(comment_list) == 1:
            return str(comment_list[0][2])
        else:
            return

    def parse_only_extr_license(self, extr_lic):
        """
        Return an ExtractedLicense object to represent a license object.
        But does not add it to the SPDXDocument model.
        Return None if failed.
        """
        # Grab all possible values
        ident = self.get_extr_license_ident(extr_lic)
        text = self.get_extr_license_text(extr_lic)
        comment = self.get_extr_lics_comment(extr_lic)
        xrefs = self.get_extr_lics_xref(extr_lic)
        name = self.get_extr_lic_name(extr_lic)

        if not ident:
            # Must have identifier
            return

        # Set fields
        # FIXME: the constructor of the license should always accept a name
        lic = license.ExtractedLicense(ident)
        if text is not None:
            lic.text = text
        if name is not None:
            lic.full_name = name
        if comment is not None:
            lic.comment = comment
        lic.cross_ref = list(map(lambda x: str(x), xrefs))
        return lic

    def handle_extracted_license(self, extr_lic):
        """
        Build and return an ExtractedLicense or None.
        Note that this function adds the license to the document.
        """
        lic = self.parse_only_extr_license(extr_lic)
        if lic is not None:
            self.doc.add_extr_lic(lic)
        return lic

    def _handle_license_list(self, lics_set, cls=None):
        """
        Return a license representing a `cls` object (LicenseConjunction
        or LicenseDisjunction) from a list of license resources or None.
        """
        licenses = []
        for _, _, lics_member in self.graph.triples(
            (lics_set, self.spdx_namespace["member"], None)
        ):
            try:
                licenses.append(self.handle_lics(lics_member))
            except CardinalityError:
                self.value_error("LICS_LIST_MEMBER", lics_member)
                break
        if len(licenses) > 1:
            return reduce(lambda a, b: cls(a, b), licenses)
        else:
            self.value_error("PKG_CONC_LIST", "")
            return

    def handle_conjunctive_list(self, lics_set):
        """
        Return a license representing the conjunction from a list of
        license resources or None.
        """
        return self._handle_license_list(lics_set, cls=license.LicenseConjunction)

    def handle_disjunctive_list(self, lics_set):
        """
        Return a license representing the disjunction from a list of
        license resources or None.
        """
        return self._handle_license_list(lics_set, cls=license.LicenseDisjunction)


class PackageParser(LicenseParser):
    """
    Helper class for parsing packages.
    """

    def __init__(self, builder, logger):
        super(PackageParser, self).__init__(builder, logger)

    def parse_package(self, p_term):
        """
        Parse package fields.
        """
        # Check there is a package name
        if not (p_term, self.spdx_namespace["name"], None) in self.graph:
            self.error = True
            self.logger.log("Package must have a name.")
            # Create dummy package so that we may continue parsing the rest of
            # the package fields.
            self.builder.create_package(self.doc, "dummy_package")
        else:
            for _s, _p, o in self.graph.triples(
                (p_term, self.spdx_namespace["name"], None)
            ):
                try:
                    self.builder.create_package(self.doc, str(o))
                except CardinalityError:
                    self.more_than_one_error("Package name")
                    break
        # Set SPDXID
        try:
            if p_term.count("#", 0, len(p_term)) == 1:
                pkg_spdx_id = p_term.split("#")[-1]
                self.builder.set_pkg_spdx_id(self.doc, pkg_spdx_id)
            else:
                self.value_error("PKG_SPDX_ID_VALUE", p_term)
        except SPDXValueError:
            self.value_error("PKG_SPDX_ID_VALUE", p_term)

        self.p_pkg_vinfo(p_term, self.spdx_namespace["versionInfo"])
        self.p_pkg_fname(p_term, self.spdx_namespace["packageFileName"])
        self.p_pkg_suppl(p_term, self.spdx_namespace["supplier"])
        self.p_pkg_originator(p_term, self.spdx_namespace["originator"])
        self.p_pkg_down_loc(p_term, self.spdx_namespace["downloadLocation"])
        self.p_pkg_files_analyzed(p_term, self.spdx_namespace["filesAnalyzed"])
        self.p_pkg_homepg(p_term, self.doap_namespace["homepage"])
        self.p_pkg_checksum(p_term, self.spdx_namespace["checksum"])
        self.p_pkg_src_info(p_term, self.spdx_namespace["sourceInfo"])
        self.p_pkg_verif_code(p_term, self.spdx_namespace["packageVerificationCode"])
        self.p_pkg_attribution_text(p_term, self.spdx_namespace["attributionText"])
        self.p_pkg_lic_conc(p_term, self.spdx_namespace["licenseConcluded"])
        self.p_pkg_lic_decl(p_term, self.spdx_namespace["licenseDeclared"])
        self.p_pkg_lics_info_from_files(
            p_term, self.spdx_namespace["licenseInfoFromFiles"]
        )
        self.p_pkg_comments_on_lics(p_term, self.spdx_namespace["licenseComments"])
        self.p_pkg_cr_text(p_term, self.spdx_namespace["copyrightText"])
        self.p_pkg_summary(p_term, self.spdx_namespace["summary"])
        self.p_pkg_descr(p_term, self.spdx_namespace["description"])
        self.p_pkg_comment(p_term, self.spdx_namespace["comment"])

    def p_pkg_cr_text(self, p_term, predicate):
        try:
            for _, _, text in self.graph.triples((p_term, predicate, None)):
                self.builder.set_pkg_cr_text(
                    self.doc, str(self.to_special_value(text))
                )
        except CardinalityError:
            self.more_than_one_error("package copyright text")

    def p_pkg_summary(self, p_term, predicate):
        try:
            for _, _, summary in self.graph.triples((p_term, predicate, None)):
                self.builder.set_pkg_summary(self.doc, str(summary))
        except CardinalityError:
            self.more_than_one_error("package summary")

    def p_pkg_descr(self, p_term, predicate):
        try:
            for _, _, desc in self.graph.triples((p_term, predicate, None)):
                self.builder.set_pkg_desc(self.doc, str(desc))
        except CardinalityError:
            self.more_than_one_error("package description")

    def p_pkg_comment(self, p_term, predicate):
        try:
            for _, _, comment in self.graph.triples((p_term, predicate, None)):
                self.builder.set_pkg_comment(self.doc, str(comment))
        except CardinalityError:
            self.more_than_one_error("package comment")

    def p_pkg_attribution_text(self, p_term, predicate):
        try:
            for _, _, attribute_text in self.graph.triples((p_term, predicate, None)):
                self.builder.set_pkg_attribution_text(
                    self.doc, str(attribute_text)
                )
        except CardinalityError:
            self.more_than_one_error("package attribution text")

    def p_pkg_comments_on_lics(self, p_term, predicate):
        for _, _, comment in self.graph.triples((p_term, predicate, None)):
            try:
                self.builder.set_pkg_license_comment(self.doc, str(comment))
            except CardinalityError:
                self.more_than_one_error("package comments on license")
                break

    def p_pkg_lics_info_from_files(self, p_term, predicate):
        for _, _, lics in self.graph.triples((p_term, predicate, None)):
            try:
                if (
                    lics,
                    RDF.type,
                    self.spdx_namespace["ExtractedLicensingInfo"],
                ) in self.graph:
                    self.builder.set_pkg_license_from_file(
                        self.doc, self.parse_only_extr_license(lics)
                    )
                else:
                    self.builder.set_pkg_license_from_file(
                        self.doc, self.handle_lics(lics)
                    )

            except SPDXValueError:
                self.value_error("PKG_LICS_INFO_FILES", lics)

    def p_pkg_lic_decl(self, p_term, predicate):
        self.handle_pkg_lic(p_term, predicate, self.builder.set_pkg_license_declared)

    def handle_pkg_lic(self, p_term, predicate, builder_func):
        """
        Handle package lics concluded or declared.
        """
        try:
            for _, _, licenses in self.graph.triples((p_term, predicate, None)):
                if (
                    licenses,
                    RDF.type,
                    self.spdx_namespace["ConjunctiveLicenseSet"],
                ) in self.graph:
                    lics = self.handle_conjunctive_list(licenses)
                    builder_func(self.doc, lics)

                elif (
                    licenses,
                    RDF.type,
                    self.spdx_namespace["DisjunctiveLicenseSet"],
                ) in self.graph:
                    lics = self.handle_disjunctive_list(licenses)
                    builder_func(self.doc, lics)

                else:
                    try:
                        lics = self.handle_lics(licenses)
                        builder_func(self.doc, lics)
                    except SPDXValueError:
                        self.value_error("PKG_SINGLE_LICS", licenses)
        except CardinalityError:
            self.more_than_one_error("package {0}".format(predicate))

    def p_pkg_lic_conc(self, p_term, predicate):
        self.handle_pkg_lic(p_term, predicate, self.builder.set_pkg_licenses_concluded)

    def p_pkg_verif_code(self, p_term, predicate):
        for _, _, verifcode in self.graph.triples((p_term, predicate, None)):
            # Parse verification code
            for _, _, code in self.graph.triples(
                (verifcode, self.spdx_namespace["packageVerificationCodeValue"], None)
            ):
                try:
                    self.builder.set_pkg_verif_code(self.doc, str(code))
                except CardinalityError:
                    self.more_than_one_error("package verification code")
                    break
            # Parse excluded file
            for _, _, filename in self.graph.triples(
                (
                    verifcode,
                    self.spdx_namespace["packageVerificationCodeExcludedFile"],
                    None,
                )
            ):
                try:
                    self.builder.set_pkg_excl_file(self.doc, str(filename))
                except CardinalityError:
                    self.more_than_one_error("package verification code excluded file")
                    break

    def p_pkg_src_info(self, p_term, predicate):
        for _, _, o in self.graph.triples((p_term, predicate, None)):
            try:
                self.builder.set_pkg_source_info(self.doc, str(o))
            except CardinalityError:
                self.more_than_one_error("package source info")
                break

    def p_pkg_checksum(self, p_term, predicate):
        for _s, _p, pkg_checksum in self.graph.triples((p_term, predicate, None)):
            for _, _, value in self.graph.triples(
                (pkg_checksum, self.spdx_namespace["checksumValue"], None)
            ):
                for _, _, algo in self.graph.triples(
                        (pkg_checksum, self.spdx_namespace["algorithm"], None)
                ):
                    algorithm_identifier = convert_rdf_checksum_algorithm(str(algo))
                    checksum = Checksum(algorithm_identifier, str(value))
                    self.builder.set_pkg_checksum(self.doc, checksum)

    def p_pkg_homepg(self, p_term, predicate):
        for _s, _p, o in self.graph.triples((p_term, predicate, None)):
            try:
                self.builder.set_pkg_home(
                    self.doc, str(self.to_special_value(o))
                )
            except CardinalityError:
                self.more_than_one_error("Package home page")
                break
            except SPDXValueError:
                self.value_error("PKG_HOME_PAGE", o)

    def p_pkg_down_loc(self, p_term, predicate):
        for _s, _p, o in self.graph.triples((p_term, predicate, None)):
            try:
                self.builder.set_pkg_down_location(
                    self.doc, str(self.to_special_value(o))
                )
            except CardinalityError:
                self.more_than_one_error("Package download location")
                break
            except SPDXValueError:
                self.value_error("PKG_DOWN_LOC", o)

    def p_pkg_files_analyzed(self, p_term, predicate):
        for _s, _p, o in self.graph.triples((p_term, predicate, None)):
            try:
                self.builder.set_pkg_files_analyzed(self.doc, str(o))
            except CardinalityError:
                self.more_than_one_error("Package Files Analyzed")
                break
            except SPDXValueError:
                self.value_error("PKG_FILES_ANALYZED_VALUE", o)

    def p_pkg_originator(self, p_term, predicate):
        for _s, _p, o in self.graph.triples((p_term, predicate, None)):
            try:
                originator = self.to_special_value(o)
                if isinstance(originator, (utils.NoAssert, utils.SPDXNone, utils.UnKnown)):
                    self.builder.set_pkg_originator(self.doc, originator)
                else:
                    ent = self.builder.create_entity(self.doc, originator)
                    self.builder.set_pkg_originator(self.doc, ent)
            except CardinalityError:
                self.more_than_one_error("Package originator")
                break
            except SPDXValueError:
                self.value_error("PKG_ORIGINATOR_VALUE", o)

    def p_pkg_suppl(self, p_term, predicate):
        for _s, _p, o in self.graph.triples((p_term, predicate, None)):
            try:
                supplier = self.to_special_value(o)
                if isinstance(supplier, (utils.NoAssert, utils.SPDXNone, utils.UnKnown)):
                    self.builder.set_pkg_supplier(self.doc, supplier)
                else:
                    ent = self.builder.create_entity(self.doc, supplier)
                    self.builder.set_pkg_supplier(self.doc, ent)
            except CardinalityError:
                self.more_than_one_error("Package supplier")
                break
            except SPDXValueError:
                self.value_error("PKG_SUPPL_VALUE", o)

    def p_pkg_fname(self, p_term, predicate):
        for _s, _p, o in self.graph.triples((p_term, predicate, None)):
            try:
                self.builder.set_pkg_file_name(self.doc, str(o))
            except CardinalityError:
                self.more_than_one_error("Package file name")
                break

    def p_pkg_vinfo(self, p_term, predicate):
        for _s, _p, o in self.graph.triples((p_term, predicate, None)):
            try:
                self.builder.set_pkg_vers(self.doc, str(o))
            except CardinalityError:
                self.more_than_one_error("Package version info")
                break


class FileParser(LicenseParser):
    """
    Helper class for parsing files.
    """

    def __init__(self, builder, logger):
        super(FileParser, self).__init__(builder, logger)

    def parse_file(self, f_term):
        if not (f_term, self.spdx_namespace["fileName"], None) in self.graph:
            self.error = True
            self.logger.log("File must have a name.")
            # Dummy name to continue
            self.builder.set_file_name(self.doc, "Dummy file")
        else:
            for _, _, name in self.graph.triples(
                (f_term, self.spdx_namespace["fileName"], None)
            ):
                self.builder.set_file_name(self.doc, str(name))

        self.p_file_spdx_id(f_term, self.spdx_namespace["File"])
        self.p_file_type(f_term, self.spdx_namespace["fileType"])
        self.p_file_checksum(f_term, self.spdx_namespace["checksum"])
        self.p_file_lic_conc(f_term, self.spdx_namespace["licenseConcluded"])
        self.p_file_lic_info(f_term, self.spdx_namespace["licenseInfoInFile"])
        self.p_file_comments_on_lics(f_term, self.spdx_namespace["licenseComments"])
        self.p_file_attribution_text(f_term, self.spdx_namespace["attributionText"])
        self.p_file_cr_text(f_term, self.spdx_namespace["copyrightText"])
        self.p_file_artifact(f_term, self.spdx_namespace["artifactOf"])
        self.p_file_comment(f_term, RDFS.comment)
        self.p_file_notice(f_term, self.spdx_namespace["noticeText"])
        self.p_file_contributor(f_term, self.spdx_namespace["fileContributor"])
        self.p_file_depends(f_term, self.spdx_namespace["fileDependency"])

    def get_file_name(self, f_term):
        """Returns first found fileName property or None if not found."""
        for _, _, name in self.graph.triples(
            (f_term, self.spdx_namespace["fileName"], None)
        ):
            return name
        return

    def p_file_depends(self, f_term, predicate):
        """
        Set file dependencies.
        """
        for _, _, other_file in self.graph.triples((f_term, predicate, None)):
            name = self.get_file_name(other_file)
            if name is not None:
                self.builder.add_file_dep(str(name))
            else:
                self.error = True
                msg = "File depends on file with no name"
                self.logger.log(msg)

    def p_file_contributor(self, f_term, predicate):
        """
        Parse all file contributors and adds them to the model.
        """
        for _, _, contributor in self.graph.triples((f_term, predicate, None)):
            self.builder.add_file_contribution(self.doc, str(contributor))

    def p_file_notice(self, f_term, predicate):
        """
        Set file notice text.
        """
        try:
            for _, _, notice in self.graph.triples((f_term, predicate, None)):
                self.builder.set_file_notice(self.doc, str(notice))
        except CardinalityError:
            self.more_than_one_error("file notice")

    def p_file_comment(self, f_term, predicate):
        """
        Set file comment text.
        """
        try:
            for _, _, comment in self.graph.triples((f_term, predicate, None)):
                self.builder.set_file_comment(self.doc, str(comment))
        except CardinalityError:
            self.more_than_one_error("file comment")

    def p_file_attribution_text(self, f_term, predicate):
        """
        Set file attribution text
        """
        try:
            for _, _, attribute_text in self.graph.triples((f_term, predicate, None)):
                self.builder.set_file_attribution_text(
                    self.doc, str(attribute_text)
                )
        except CardinalityError:
            self.more_than_one_error("file attribution text")

    def p_file_artifact(self, f_term, predicate):
        """
        Handle file artifactOf.
        Note: does not handle artifact of project URI.
        """
        for _, _, project in self.graph.triples((f_term, predicate, None)):
            if (project, RDF.type, self.doap_namespace["Project"]):
                self.p_file_project(project)
            else:
                self.error = True
                msg = "File must be artifact of doap:Project"
                self.logger.log(msg)

    def p_file_project(self, project):
        """
        Helper function for parsing doap:project name and homepage.
        and setting them using the file builder.
        """
        for _, _, name in self.graph.triples(
            (project, self.doap_namespace["name"], None)
        ):
            self.builder.set_file_atrificat_of_project(
                self.doc, "name", str(name)
            )
        for _, _, homepage in self.graph.triples(
            (project, self.doap_namespace["homepage"], None)
        ):
            self.builder.set_file_atrificat_of_project(
                self.doc, "home", str(homepage)
            )

    def p_file_cr_text(self, f_term, predicate):
        """
        Set file copyright text.
        """
        try:
            for _, _, cr_text in self.graph.triples((f_term, predicate, None)):
                self.builder.set_file_copyright(self.doc, str(cr_text))
        except CardinalityError:
            self.more_than_one_error("file copyright text")

    def p_file_comments_on_lics(self, f_term, predicate):
        """
        Set file license comment.
        """
        try:
            for _, _, comment in self.graph.triples((f_term, predicate, None)):
                self.builder.set_file_license_comment(self.doc, str(comment))
        except CardinalityError:
            self.more_than_one_error("file comments on license")

    def p_file_lic_info(self, f_term, predicate):
        """
        Set file license information.
        """
        for _, _, info in self.graph.triples((f_term, predicate, None)):
            lic = self.handle_lics(info)
            if lic is not None:
                self.builder.set_file_license_in_file(self.doc, lic)

    def p_file_spdx_id(self, f_term, predicate):
        try:
            try:
                self.builder.set_file_spdx_id(self.doc, str(f_term))
            except SPDXValueError:
                self.value_error("FILE_SPDX_ID_VALUE", f_term)
        except CardinalityError:
            self.more_than_one_error("FILE_SPDX_ID_VALUE")

    def p_file_type(self, f_term, predicate):
        """
        Set file type.
        """
        try:
            for _, _, ftype in self.graph.triples((f_term, predicate, None)):
                try:
                    self.builder.set_file_type(self.doc, ftype)
                except SPDXValueError:
                    self.value_error("FILE_TYPE", ftype)
        except CardinalityError:
            self.more_than_one_error("file type")

    def p_file_checksum(self, f_term, predicate):
        """
        Set file checksum.
        """
        for _s, _p, file_checksum in self.graph.triples((f_term, predicate, None)):
            for _, _, value in self.graph.triples(
                (file_checksum, self.spdx_namespace["checksumValue"], None)
            ):
                for _, _, algo in self.graph.triples(
                        (file_checksum, self.spdx_namespace["algorithm"], None)
                ):
                    algorithm_identifier = convert_rdf_checksum_algorithm(str(algo))
                    checksum = Checksum(algorithm_identifier, str(value))
                    self.builder.set_file_checksum(self.doc, checksum)

    def p_file_lic_conc(self, f_term, predicate):
        """
        Set file licenses concluded.
        """
        try:
            for _, _, licenses in self.graph.triples((f_term, predicate, None)):
                if (
                    licenses,
                    RDF.type,
                    self.spdx_namespace["ConjunctiveLicenseSet"],
                ) in self.graph:
                    lics = self.handle_conjunctive_list(licenses)
                    self.builder.set_concluded_license(self.doc, lics)

                elif (
                    licenses,
                    RDF.type,
                    self.spdx_namespace["DisjunctiveLicenseSet"],
                ) in self.graph:
                    lics = self.handle_disjunctive_list(licenses)
                    self.builder.set_concluded_license(self.doc, lics)

                else:
                    try:
                        lics = self.handle_lics(licenses)
                        self.builder.set_concluded_license(self.doc, lics)
                    except SPDXValueError:
                        self.value_error("FILE_SINGLE_LICS", licenses)
        except CardinalityError:
            self.more_than_one_error("file {0}".format(predicate))


class SnippetParser(LicenseParser):
    """
    Helper class for parsing snippet information.
    """

    def __init__(self, builder, logger):
        super(SnippetParser, self).__init__(builder, logger)

    def parse_snippet(self, snippet_term):
        try:
            self.builder.create_snippet(self.doc, snippet_term)
        except SPDXValueError:
            self.value_error("SNIPPET_SPDX_ID_VALUE", snippet_term)

        for _s, _p, o in self.graph.triples(
            (snippet_term, self.spdx_namespace["name"], None)
        ):
            try:
                self.builder.set_snippet_name(self.doc, str(o))
            except CardinalityError:
                self.more_than_one_error("snippetName")
                break

        for _s, _p, o in self.graph.triples(
            (snippet_term, self.spdx_namespace["licenseComments"], None)
        ):
            try:
                self.builder.set_snippet_lic_comment(self.doc, str(o))
            except CardinalityError:
                self.more_than_one_error("licenseComments")
                break

        for _s, _p, o in self.graph.triples((snippet_term, RDFS.comment, None)):
            try:
                self.builder.set_snippet_comment(self.doc, str(o))
            except CardinalityError:
                self.more_than_one_error("comment")
                break

        for _s, _p, o in self.graph.triples(
            (snippet_term, self.spdx_namespace["copyrightText"], None)
        ):
            try:
                self.builder.set_snippet_copyright(
                    self.doc, self.to_special_value(str(o))
                )
            except CardinalityError:
                self.more_than_one_error("copyrightText")
                break

        try:
            for _, _, licenses in self.graph.triples(
                (snippet_term, self.spdx_namespace["licenseConcluded"], None)
            ):
                if (
                    licenses,
                    RDF.type,
                    self.spdx_namespace["ConjunctiveLicenseSet"],
                ) in self.graph:
                    lics = self.handle_conjunctive_list(licenses)
                    self.builder.set_snip_concluded_license(self.doc, lics)

                elif (
                    licenses,
                    RDF.type,
                    self.spdx_namespace["DisjunctiveLicenseSet"],
                ) in self.graph:
                    lics = self.handle_disjunctive_list(licenses)
                    self.builder.set_snip_concluded_license(self.doc, lics)

                else:
                    try:
                        lics = self.handle_lics(licenses)
                        self.builder.set_snip_concluded_license(self.doc, lics)
                    except SPDXValueError:
                        self.value_error("SNIPPET_CONCLUDED_LICENSE", licenses)
        except CardinalityError:
            self.more_than_one_error(
                "package {0}".format(self.spdx_namespace["licenseConcluded"])
            )

        for _, _, info in self.graph.triples(
            (snippet_term, self.spdx_namespace["licenseInfoInSnippet"], None)
        ):
            lic = self.handle_lics(info)
            if lic is not None:
                try:
                    self.builder.set_snippet_lics_info(self.doc, lic)
                except SPDXValueError:
                    self.value_error("SNIPPET_LIC_INFO", lic)

        for _s, _p, o in self.graph.triples(
            (snippet_term, self.spdx_namespace["snippetFromFile"], None)
        ):
            try:
                self.builder.set_snip_from_file_spdxid(self.doc, str(o))
            except CardinalityError:
                self.more_than_one_error("snippetFromFile")
                break

        try:
            for _, _, attribute_text in self.graph.triples(
                (snippet_term, self.spdx_namespace["attributionText"], None)
            ):
                self.builder.set_snippet_attribution_text(
                    self.doc, str(attribute_text)
                )
        except CardinalityError:
            self.more_than_one_error("snippetAttributionText")


class ReviewParser(BaseParser):
    """
    Helper class for parsing review information.
    """

    def __init__(self, builder, logger):
        super(ReviewParser, self).__init__(builder, logger)

    def parse_review(self, r_term):
        reviewer = self.get_reviewer(r_term)
        reviewed_date = self.get_review_date(r_term)
        if reviewer is not None:
            self.builder.add_reviewer(self.doc, reviewer)
            if reviewed_date is not None:
                try:
                    self.builder.add_review_date(self.doc, reviewed_date)
                except SPDXValueError:
                    self.value_error("REVIEW_DATE", reviewed_date)
            comment = self.get_review_comment(r_term)
            if comment is not None:
                self.builder.add_review_comment(self.doc, comment)

    def get_review_comment(self, r_term):
        """
        Return review comment or None if found none or more than one.
        Report errors.
        """
        comment_list = list(self.graph.triples((r_term, RDFS.comment, None)))
        if len(comment_list) > 1:
            self.error = True
            msg = "Review can have at most one comment"
            self.logger.log(msg)
            return
        else:
            return str(comment_list[0][2])

    def get_review_date(self, r_term):
        """
        Return review date or None if not found.
        Report error on failure.
        Note does not check value format.
        """
        reviewed_list = list(
            self.graph.triples((r_term, self.spdx_namespace["reviewDate"], None))
        )
        if len(reviewed_list) != 1:
            self.error = True
            msg = "Review must have exactly one review date"
            self.logger.log(msg)
            return
        return str(reviewed_list[0][2])

    def get_reviewer(self, r_term):
        """
        Return reviewer as creator object or None if failed.
        Report errors on failure.
        """
        reviewer_list = list(
            self.graph.triples((r_term, self.spdx_namespace["reviewer"], None))
        )
        if len(reviewer_list) != 1:
            self.error = True
            msg = "Review must have exactly one reviewer"
            self.logger.log(msg)
            return
        try:
            return self.builder.create_entity(
                self.doc, str(reviewer_list[0][2])
            )
        except SPDXValueError:
            self.value_error("REVIEWER_VALUE", reviewer_list[0][2])


class AnnotationParser(BaseParser):
    """
    Helper class for parsing annotation information.
    """

    def __init__(self, builder, logger):
        super(AnnotationParser, self).__init__(builder, logger)

    def parse_annotation(self, r_term):
        annotator = self.get_annotator(r_term)
        annotation_date = self.get_annotation_date(r_term)
        if annotator is not None:
            self.builder.add_annotator(self.doc, annotator)
            if annotation_date is not None:
                try:
                    self.builder.add_annotation_date(self.doc, annotation_date)
                except SPDXValueError:
                    self.value_error("ANNOTATION_DATE", annotation_date)
            comment = self.get_annotation_comment(r_term)
            if comment is not None:
                self.builder.add_annotation_comment(self.doc, comment)
            annotation_type = self.get_annotation_type(r_term)
            self.builder.add_annotation_type(self.doc, annotation_type)
            try:
                self.builder.set_annotation_spdx_id(self.doc, str(r_term))
            except CardinalityError:
                self.more_than_one_error("SPDX Identifier Reference")

    def get_annotation_type(self, r_term):
        """
        Return annotation type or None if found none or more than one.
        Report errors on failure.
        """
        for _, _, typ in self.graph.triples(
            (r_term, self.spdx_namespace["annotationType"], None)
        ):
            if typ is not None:
                return str(typ)
            else:
                self.error = True
                msg = "Annotation must have exactly one annotation type."
                self.logger.log(msg)
                return

    def get_annotation_comment(self, r_term):
        """
        Return annotation comment or None if found none or more than one.
        Report errors.
        """
        comment_list = list(self.graph.triples((r_term, RDFS.comment, None)))
        if len(comment_list) > 1:
            self.error = True
            msg = "Annotation can have at most one comment."
            self.logger.log(msg)
            return
        else:
            return str(comment_list[0][2])

    def get_annotation_date(self, r_term):
        """
        Return annotation date or None if not found.
        Report error on failure.
        Note does not check value format.
        """
        annotation_date_list = list(
            self.graph.triples((r_term, self.spdx_namespace["annotationDate"], None))
        )
        if len(annotation_date_list) != 1:
            self.error = True
            msg = "Annotation must have exactly one annotation date."
            self.logger.log(msg)
            return
        return str(annotation_date_list[0][2])

    def get_annotator(self, r_term):
        """
        Return annotator as creator object or None if failed.
        Report errors on failure.
        """
        annotator_list = list(
            self.graph.triples((r_term, self.spdx_namespace["annotator"], None))
        )
        if len(annotator_list) != 1:
            self.error = True
            msg = "Annotation must have exactly one annotator"
            self.logger.log(msg)
            return
        try:
            return self.builder.create_entity(
                self.doc, str(annotator_list[0][2])
            )
        except SPDXValueError:
            self.value_error("ANNOTATOR_VALUE", annotator_list[0][2])


class RelationshipParser(BaseParser):
    """
    Helper Class for parsing relationship information
    """

    def __init__(self, builder, logger):
        super(RelationshipParser, self).__init__(builder, logger)

    def parse_relationship(self, subject_term, relation_term):
        relationship = self.get_relationship(subject_term, relation_term)
        relationship_comment = self.get_relationship_comment(relation_term)
        if relationship is not None:
            relationship_added: bool = self.builder.add_relationship(self.doc, relationship)
            if relationship_comment is not None and relationship_added:
                self.builder.add_relationship_comment(self.doc, relationship_comment)

    def get_relationship(self, subject_term, relation_term):
        """
        Returns a string with relationship type and the related elements.
        """
        relation_subject = str(subject_term.split("#")[1])

        for _, _, rtype in self.graph.triples(
            (relation_term, self.spdx_namespace["relationshipType"], None)
        ):
            try:
                if rtype.endswith("describes"):
                    rtype = "DESCRIBES"
                elif rtype.endswith("describedBy"):
                    rtype = "DESCRIBED_BY"
                elif rtype.endswith("contains"):
                    rtype = "CONTAINS"
                elif rtype.endswith("containedBy"):
                    rtype = "CONTAINED_BY"
                elif rtype.endswith("dependsOn"):
                    rtype = "DEPENDS_ON"
                elif rtype.endswith("dependencyOf"):
                    rtype = "DEPENDENCY_OF"
                elif rtype.endswith("dependencyManifestOf"):
                    rtype = "DEPENDENCY_MANIFEST_OF"
                elif rtype.endswith("buildDependencyOf"):
                    rtype = "BUILD_DEPENDENCY_OF"
                elif rtype.endswith("devDependencyOf"):
                    rtype = "DEV_DEPENDENCY_OF"
                elif rtype.endswith("optionalDependencyOf"):
                    rtype = "OPTIONAL_DEPENDENCY_OF"
                elif rtype.endswith("providedDependencyOf"):
                    rtype = "PROVIDED_DEPENDENCY_OF"
                elif rtype.endswith("testDependencyOf"):
                    rtype = "TEST_DEPENDENCY_OF"
                elif rtype.endswith("runtimeDependencyOf"):
                    rtype = "RUNTIME_DEPENDENCY_OF"
                elif rtype.endswith("exampleOf"):
                    rtype = "EXAMPLE_OF"
                elif rtype.endswith("generates"):
                    rtype = "GENERATES"
                elif rtype.endswith("generatedFrom"):
                    rtype = "GENERATED_FROM"
                elif rtype.endswith("ancestorOf"):
                    rtype = "ANCESTOR_OF"
                elif rtype.endswith("descendantOf"):
                    rtype = "DESCENDANT_OF"
                elif rtype.endswith("variantOf"):
                    rtype = "VARIANT_OF"
                elif rtype.endswith("distributionArtifact"):
                    rtype = "DISTRIBUTION_ARTIFACT"
                elif rtype.endswith("patchFor"):
                    rtype = "PATCH_FOR"
                elif rtype.endswith("patchApplied"):
                    rtype = "PATCH_APPLIED"
                elif rtype.endswith("copyOf"):
                    rtype = "COPY_OF"
                elif rtype.endswith("fileAdded"):
                    rtype = "FILE_ADDED"
                elif rtype.endswith("fileDeleted"):
                    rtype = "FILE_DELETED"
                elif rtype.endswith("fileModified"):
                    rtype = "FILE_MODIFIED"
                elif rtype.endswith("expandedFromArchive"):
                    rtype = "EXPANDED_FROM_ARCHIVE"
                elif rtype.endswith("dynamicLink"):
                    rtype = "DYNAMIC_LINK"
                elif rtype.endswith("staticLink"):
                    rtype = "STATIC_LINK"
                elif rtype.endswith("dataFileOf"):
                    rtype = "DATA_FILE_OF"
                elif rtype.endswith("testCaseOf"):
                    rtype = "TEST_CASE_OF"
                elif rtype.endswith("buildToolOf"):
                    rtype = "BUILD_TOOL_OF"
                elif rtype.endswith("devToolOf"):
                    rtype = "DEV_TOOL_OF"
                elif rtype.endswith("testOf"):
                    rtype = "TEST_OF"
                elif rtype.endswith("testToolOf"):
                    rtype = "TEST_TOOL_OF"
                elif rtype.endswith("documentationOf"):
                    rtype = "DOCUMENTATION_OF"
                elif rtype.endswith("optionalComponentOf"):
                    rtype = "OPTIONAL_COMPONENT_OF"
                elif rtype.endswith("metafileOf"):
                    rtype = "METAFILE_OF"
                elif rtype.endswith("packageOf"):
                    rtype = "PACKAGE_OF"
                elif rtype.endswith("amends"):
                    rtype = "AMENDS"
                elif rtype.endswith("prerequisiteFor"):
                    rtype = "PREREQUISITE_FOR"
                elif rtype.endswith("hasPrerequisite"):
                    rtype = "HAS_PREREQUISITE"
                elif rtype.endswith("other"):
                    rtype = "OTHER"
                elif rtype.endswith("specificationFor"):
                    rtype = "SPECIFICATION_FOR"
                elif rtype.endswith("requirementDescriptionFor"):
                    rtype = "REQUIREMENT_DESCRIPTION_FOR"

            except SPDXValueError:
                self.value_error("RELATIONSHIP", rtype)

        try:
            for sub, pre, rel_ele in self.graph.triples(
                (relation_term, self.spdx_namespace["relatedSpdxElement"], None)
            ):
                related_element = str(rel_ele.split("#")[1]) if '#' in rel_ele else rel_ele
        except:
            related_element = None

        try:
            if related_element == None:
                return str(relation_subject + " " + rtype)
            else:
                return str(
                    relation_subject + " " + rtype + " " + related_element
                )

        except SPDXValueError:
            self.value_error("RELATIONSHIP_VALUE", relation_subject + " " + rtype)

    def get_relationship_comment(self, relation_term):
        """
        Returns relationship comment or None if found none or more than one.
        Reports errors.
        """

        comment_list = list(self.graph.triples((relation_term, RDFS.comment, None)))
        if len(comment_list) == 0:
            return None
        else:
            if len(comment_list) > 1:
                self.error = True
                msg = "Relationship can have at most one comment."
                self.logger.log(msg)
                return
            else:
                return str(comment_list[0][2])


class Parser(
    PackageParser,
    FileParser,
    SnippetParser,
    ReviewParser,
    AnnotationParser,
    RelationshipParser,
):
    """
    RDF/XML file parser.
    """

    def __init__(self, builder, logger):
        super(Parser, self).__init__(builder, logger)

    def parse(self, fil):
        """
        Parse a file and returns a document object.
        fil is a file like object.
        """
        self.error = False
        self.graph = Graph()
        self.graph.parse(file=fil, format="xml")
        self.doc = document.Document()

        for s, _p, o in self.graph.triples(
            (None, RDF.type, self.spdx_namespace["SpdxDocument"])
        ):
            self.parse_doc_fields(s)

        for s, _p, o in self.graph.triples(
            (None, RDF.type, self.spdx_namespace["ExternalDocumentRef"])
        ):
            self.parse_ext_doc_ref(s)

        for s, _p, o in self.graph.triples(
            (None, RDF.type, self.spdx_namespace["CreationInfo"])
        ):
            self.parse_creation_info(s)

        for s, _p, o in self.graph.triples(
            (None, None, self.spdx_namespace["ExtractedLicensingInfo"])
        ):
            self.handle_extracted_license(s)

        for s, _p, o in self.graph.triples(
            (None, RDF.type, self.spdx_namespace["Package"])
        ):
            self.parse_package(s)

        for s, _p, o in self.graph.triples(
            (None, RDF.type, self.spdx_namespace["ExternalRef"])
        ):
            self.parse_pkg_ext_ref(s)

        for s, _p, o in self.graph.triples(
            (None, self.spdx_namespace["referencesFile"], None)
        ):
            self.parse_file(o)

        for s, _p, o in self.graph.triples(
            (None, RDF.type, self.spdx_namespace["Snippet"])
        ):
            self.parse_snippet(s)

        for s, _p, o in self.graph.triples(
            (None, self.spdx_namespace["reviewed"], None)
        ):
            self.parse_review(o)

        for s, _p, o in self.graph.triples(
            (None, self.spdx_namespace["annotation"], None)
        ):
            self.parse_annotation(o)

        for s, _p, o in self.graph.triples(
            (None, self.spdx_namespace["relationship"], None)
        ):
            self.parse_relationship(s, o)

        validation_messages = ErrorMessages()
        # Report extra errors if self.error is False otherwise there will be
        # redundant messages
        validation_messages = self.doc.validate(validation_messages)
        if not self.error:
            if validation_messages:
                for msg in validation_messages:
                    self.logger.log(msg)
                self.error = True
        return self.doc, self.error

    def parse_creation_info(self, ci_term):
        """
        Parse creators, created and comment.
        """
        for _s, _p, o in self.graph.triples(
            (ci_term, self.spdx_namespace["creator"], None)
        ):
            try:
                ent = self.builder.create_entity(self.doc, str(o))
                self.builder.add_creator(self.doc, ent)
            except SPDXValueError:
                self.value_error("CREATOR_VALUE", o)

        for _s, _p, o in self.graph.triples(
            (ci_term, self.spdx_namespace["created"], None)
        ):
            try:
                self.builder.set_created_date(self.doc, str(o))
            except SPDXValueError:
                self.value_error("CREATED_VALUE", o)
            except CardinalityError:
                self.more_than_one_error("created")
                break

        for _s, _p, o in self.graph.triples((ci_term, RDFS.comment, None)):
            try:
                self.builder.set_creation_comment(self.doc, str(o))
            except CardinalityError:
                self.more_than_one_error("CreationInfo comment")
                break
        for _s, _p, o in self.graph.triples(
            (ci_term, self.spdx_namespace["licenseListVersion"], None)
        ):
            try:
                self.builder.set_lics_list_ver(self.doc, str(o))
            except CardinalityError:
                self.more_than_one_error("licenseListVersion")
                break
            except SPDXValueError:
                self.value_error("LL_VALUE", o)

    def parse_doc_fields(self, doc_term):
        """
        Parse the version, data license, name, SPDX Identifier, namespace,
        and comment.
        """
        try:
            self.builder.set_doc_spdx_id(self.doc, str(doc_term))
        except SPDXValueError:
            self.value_error("DOC_SPDX_ID_VALUE", doc_term)
        try:
            if doc_term.count("#", 0, len(doc_term)) <= 1:
                doc_namespace = doc_term.split("#")[0]
                self.builder.set_doc_namespace(self.doc, doc_namespace)
            else:
                self.value_error("DOC_NAMESPACE_VALUE", doc_term)
        except SPDXValueError:
            self.value_error("DOC_NAMESPACE_VALUE", doc_term)
        for _s, _p, o in self.graph.triples(
            (doc_term, self.spdx_namespace["specVersion"], None)
        ):
            try:
                self.builder.set_doc_version(self.doc, str(o))
            except SPDXValueError:
                self.value_error("DOC_VERS_VALUE", o)
            except CardinalityError:
                self.more_than_one_error("specVersion")
                break
        for _s, _p, o in self.graph.triples(
            (doc_term, self.spdx_namespace["dataLicense"], None)
        ):
            try:
                self.builder.set_doc_data_lic(self.doc, str(o))
            except SPDXValueError:
                self.value_error("DOC_D_LICS", o)
            except CardinalityError:
                self.more_than_one_error("dataLicense")
                break
        for _s, _p, o in self.graph.triples(
            (doc_term, self.spdx_namespace["name"], None)
        ):
            try:
                self.builder.set_doc_name(self.doc, str(o))
            except CardinalityError:
                self.more_than_one_error("name")
                break
        for _s, _p, o in self.graph.triples((doc_term, RDFS.comment, None)):
            try:
                self.builder.set_doc_comment(self.doc, str(o))
            except CardinalityError:
                self.more_than_one_error("Document comment")
                break

    def parse_ext_doc_ref(self, ext_doc_ref_term):
        """
        Parse the External Document ID, SPDX Document URI and Checksum.
        """
        for _s, _p, o in self.graph.triples(
            (ext_doc_ref_term, self.spdx_namespace["externalDocumentId"], None)
        ):
            try:
                self.builder.set_ext_doc_id(self.doc, str(o))
            except SPDXValueError:
                self.value_error("EXT_DOC_REF_VALUE", "External Document ID")
                break

        for _s, _p, o in self.graph.triples(
            (ext_doc_ref_term, self.spdx_namespace["spdxDocument"], None)
        ):
            try:
                self.builder.set_spdx_doc_uri(self.doc, str(o))
            except SPDXValueError:
                self.value_error("EXT_DOC_REF_VALUE", "SPDX Document URI")
                break

        for _s, _p, checksum in self.graph.triples(
            (ext_doc_ref_term, self.spdx_namespace["checksum"], None)
        ):
            for _, _, value in self.graph.triples(
                (checksum, self.spdx_namespace["checksumValue"], None)
            ):
                try:
                    self.builder.set_chksum(self.doc, str(value))
                except SPDXValueError:
                    self.value_error("EXT_DOC_REF_VALUE", "Checksum")
                    break

    def parse_pkg_ext_ref(self, pkg_ext_term):
        """
        Parse the category, type, locator, and comment.
        """
        for _s, _p, o in self.graph.triples(
            (pkg_ext_term, self.spdx_namespace["referenceCategory"], None)
        ):
            try:
                self.builder.set_pkg_ext_ref_category(self.doc, str(o))
            except SPDXValueError:
                self.value_error(
                    "PKG_EXT_REF_CATEGORY", "Package External Reference Category"
                )
                break

        for _s, _p, o in self.graph.triples(
            (pkg_ext_term, self.spdx_namespace["referenceType"], None)
        ):
            try:
                self.builder.set_pkg_ext_ref_type(self.doc, str(o))
            except SPDXValueError:
                self.value_error("PKG_EXT_REF_TYPE", "Package External Reference Type")
                break

        for _s, _p, o in self.graph.triples(
            (pkg_ext_term, self.spdx_namespace["referenceLocator"], None)
        ):
            self.builder.set_pkg_ext_ref_locator(self.doc, str(o))

        for _s, _p, o in self.graph.triples((pkg_ext_term, RDFS.comment, None)):
            try:
                self.builder.set_pkg_ext_ref_comment(self.doc, str(o))
            except CardinalityError:
                self.more_than_one_error("Package External Reference Comment")
                break



# Subset of nose tools (the parts included have not been modified)
# Borrowed from:
# https://raw.githubusercontent.com/nose-devs/nose/7c26ad1e6b7d308cafa328ad34736d34028c122a/nose/tools/nontrivial.py
# Copyright (c) 2005-2009 Jason Pellerin and others.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# This program is free software; you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St,
# Fifth Floor, Boston, MA 02110-1301 USA
#

"""Tools not exempt from being descended into in tracebacks"""

def make_decorator(func):
    """
    Wraps a test decorator so as to properly replicate metadata
    of the decorated function, including nose's additional stuff
    (namely, setup and teardown).
    """
    def decorate(newfunc):
        if hasattr(func, 'compat_func_name'):
            name = func.compat_func_name
        else:
            name = func.__name__
        newfunc.__dict__ = func.__dict__
        newfunc.__doc__ = func.__doc__
        newfunc.__module__ = func.__module__
        if not hasattr(newfunc, 'compat_co_firstlineno'):
            newfunc.compat_co_firstlineno = func.__code__.co_firstlineno
        try:
            newfunc.__name__ = name
        except TypeError:
            # can't set func name in 2.3
            newfunc.compat_func_name = name
        return newfunc
    return decorate


def raises(*exceptions):
    """Test must raise one of expected exceptions to pass.

    Example use::

      @raises(TypeError, ValueError)
      def test_raises_type_error():
          raise TypeError("This test passes")

      @raises(Exception)
      def test_that_fails_by_passing():
          pass

    If you want to test many assertions about exceptions in a single test,
    you may want to use `assert_raises` instead.
    """
    valid = ' or '.join([e.__name__ for e in exceptions])
    def decorate(func):
        name = func.__name__
        def newfunc(*arg, **kw):
            try:
                func(*arg, **kw)
            except exceptions:
                pass
            except:
                raise
            else:
                message = "%s() did not raise %s" % (name, valid)
                raise AssertionError(message)
        newfunc = make_decorator(func)(newfunc)
        return newfunc
    return decorate




#!/usr/bin/env python

# Parses an RDF file and writes it out pretty-printed.
# Usage: pp_rdf <infile> <outfile>

if __name__ == "__main__":
    import sys
    from spdx.parsers.rdf import Parser
    from spdx.parsers.loggers import StandardLogger
    from spdx.parsers.rdfbuilders import Builder
    from spdx.writers.rdf import write_document

    infile = sys.argv[1]
    outfile = sys.argv[2]
    p = Parser(Builder(), StandardLogger())
    with open(infile) as f:
        doc, error = p.parse(f)
        if not error:
            with open(outfile, mode="wb") as out:
                write_document(doc, out)

        else:
            print("Errors while parsing")
